from terasim.agent.agent_sensor import AgentSensor
from terasim.overlay import traci


class EgoSensor(AgentSensor):
    """A sensor for reporting basic states (position, speed, heading, etc.)"""

    DEFAULT_PARAMS = dict(
        fields={
            "velocity": traci.vehicle.getSpeed,
            "position": traci.vehicle.getPosition,
            "position3d": traci.vehicle.getPosition3D,
            "heading": traci.vehicle.getAngle,
            "edge_id": traci.vehicle.getRoadID,
            "lane_id": traci.vehicle.getLaneID,
            "lane_index": traci.vehicle.getLaneIndex,
            "acceleration": traci.vehicle.getAcceleration,
        }
    )

    def __init__(self, name="ego", **params):
        """Initialize the ego state sensor for the vehicle.

        Args:
            name (str, optional): The name of the sensor. Defaults to "ego".
            params (dict, optional): The parameters of the sensor.
        """
        super().__init__(name, **params)
        self._length = None
        self._width = None
        self._height = None

    @property
    def length(self):
        """Get the length of the vehicle.

        Returns:
            float: The length of the vehicle.
        """
        if self._length is None:
            self._length = traci.vehicle.getLength(self._agent.id)
        return self._length

    @property
    def width(self):
        """Get the width of the vehicle.

        Returns:
            float: The width of the vehicle.
        """
        if self._width is None:
            self._width = traci.vehicle.getWidth(self._agent.id)
        return self._width

    @property
    def height(self):
        """Get the height of the vehicle.

        Returns:
            float: The height of the vehicle.
        """
        if self._height is None:
            self._height = traci.vehicle.getHeight(self._agent.id)
        return self._height

    def fetch(self) -> dict:
        """Fetch the vehicle state.

        Returns:
            dict: The vehicle state.
        """
        veh_id = self._agent.id
        data = {"veh_id": veh_id}
        for field, getter in self.params.fields.items():
            data[field] = getter(veh_id)
        data["length"] = self.length
        data["width"] = self.width
        data["height"] = self.height
        return data



from abc import ABC, abstractmethod


class VehicleFactory(ABC):
    """
    Basic Vehicle factory class to help build a vehicle, each vehicle will contain three major components: sensors, controllers, and decision models.
    Each user who would like to build a customized vehicle should build a son-class and overwrite the vehicle creation method.
    """

    @abstractmethod
    def create_vehicle(self, veh_id, simulator):
        """Create a vehicle with the given id and simulator.

        Args:
            veh_id (str): The vehicle id.
            simulator (Simulator): The simulator object.

        Returns:
            Vehicle: The vehicle object.
        """
        raise NotImplementedError("Create vehicle method not implemented!")



from abc import ABC, abstractmethod


class TrafficLightFactory(ABC):
    """
    Basic traffic light factory class to help build a traffic light, each traffic light will contain three major components: sensors, controllers, and powertrains.
    Each user who would like to build a customized traffic light should build a son-class and overwrite the traffic light creation method.
    """

    @abstractmethod
    def create_traffic_light(self, tls_id, simulator):
        """Create a traffic light with the given id and simulator.

        Args:
            tls_id (str): The traffic light id.
            simulator (Simulator): The simulator object.

        Returns:
            TrafficLight: The traffic light object.
        """
        raise NotImplementedError("Create traffic light method not implemented!")



from __future__ import division, print_function

import numpy as np
from scipy import stats

# Longitudinal policy parameters
from terasim.overlay import traci
from terasim.vehicle.decision_models.highway_base_decision_model import HighwayBaseDecisionModel

# Lateral policy parameters
POLITENESS = 0.0  # in [0, 1]
LANE_CHANGE_MIN_ACC_GAIN = 0.1  # [m/s2]
LANE_CHANGE_MAX_BRAKING_IMPOSED = 4.0  # [m/s2]

# NDD Vehicle IDM parameters
COMFORT_ACC_MAX = 2  # [m/s2]
COMFORT_ACC_MIN = -4.0  # [m/s2]
DISTANCE_WANTED = 5.0  # [m]
TIME_WANTED = 1.5  # [s]
DESIRED_VELOCITY = 35  # [m/s]
DELTA = 4.0  # []

acc_low = -4
acc_high = 2
stochastic_IDM_resolution = 0.2
stochastic_IDM_prob_threshold = 1e-10
LENGTH = 5


class IDMModel(HighwayBaseDecisionModel):
    """A vehicle using both a longitudinal and a lateral decision policies.

    - Longitudinal: the IDM model computes an acceleration given the preceding vehicle's distance and velocity.
    - Lateral: the MOBIL model decides when to change lane by maximizing the acceleration of nearby vehicles.
    """

    def __init__(
        self,
        MOBIL_lc_flag=True,
        stochastic_acc_flag=False,
        IDM_parameters=None,
        MOBIL_parameters=None,
    ):
        """Initialization of the IDM mdoel

        Args:
            MOBIL_lc_flag (bool, optional): whether to include MOBIL lane change model, if not, will call SUMO lane change model. Defaults to True.
            stochastic_acc_flag (bool, optional): Whether the acceleration will be stochastic and random sampled. Defaults to False.
        """
        super().__init__()
        self.MOBIL_lc_flag = MOBIL_lc_flag
        self.stochastic_acc_flag = stochastic_acc_flag
        self.load_parameters(IDM_parameters, MOBIL_parameters)

    def load_parameters(self, IDM_parameters, MOBIL_parameters):
        """Load the IDM and MOBIL parameters.

        Args:
            IDM_parameters (dict, optional): IDM parameters. Defaults to None.
            MOBIL_parameters (dict, optional): MOBIL parameters. Defaults to None.
        """
        IDM_parameters = IDM_parameters if IDM_parameters else {}

        self.COMFORT_ACC_MAX = (
            IDM_parameters["COMFORT_ACC_MAX"]
            if "COMFORT_ACC_MAX" in IDM_parameters
            else COMFORT_ACC_MAX
        )
        self.COMFORT_ACC_MIN = (
            IDM_parameters["COMFORT_ACC_MIN"]
            if "COMFORT_ACC_MIN" in IDM_parameters
            else COMFORT_ACC_MIN
        )
        self.DISTANCE_WANTED = (
            IDM_parameters["DISTANCE_WANTED"]
            if "DISTANCE_WANTED" in IDM_parameters
            else DISTANCE_WANTED
        )
        self.TIME_WANTED = (
            IDM_parameters["TIME_WANTED"] if "TIME_WANTED" in IDM_parameters else TIME_WANTED
        )
        self.DESIRED_VELOCITY = (
            IDM_parameters["DESIRED_VELOCITY"]
            if "DESIRED_VELOCITY" in IDM_parameters
            else DESIRED_VELOCITY
        )
        self.DELTA = IDM_parameters["DELTA"] if "DELTA" in IDM_parameters else DELTA
        self.stochastic_IDM_resolution = (
            IDM_parameters["stochastic_IDM_resolution"]
            if "stochastic_IDM_resolution" in IDM_parameters
            else stochastic_IDM_resolution
        )
        self.stochastic_IDM_prob_threshold = (
            IDM_parameters["stochastic_IDM_prob_threshold"]
            if "stochastic_IDM_prob_threshold" in IDM_parameters
            else stochastic_IDM_prob_threshold
        )
        self.LENGTH = IDM_parameters["LENGTH"] if "LENGTH" in IDM_parameters else LENGTH
        self.acc_low = IDM_parameters["acc_low"] if "acc_low" in IDM_parameters else acc_low
        self.acc_high = IDM_parameters["acc_high"] if "acc_high" in IDM_parameters else acc_high

        MOBIL_parameters = MOBIL_parameters if MOBIL_parameters else {}
        self.POLITENESS = (
            MOBIL_parameters["POLITENESS"] if "POLITENESS" in MOBIL_parameters else POLITENESS
        )
        self.LANE_CHANGE_MIN_ACC_GAIN = (
            MOBIL_parameters["LANE_CHANGE_MIN_ACC_GAIN"]
            if "LANE_CHANGE_MIN_ACC_GAIN" in MOBIL_parameters
            else LANE_CHANGE_MIN_ACC_GAIN
        )
        self.LANE_CHANGE_MAX_BRAKING_IMPOSED = (
            MOBIL_parameters["LANE_CHANGE_MAX_BRAKING_IMPOSED"]
            if "LANE_CHANGE_MAX_BRAKING_IMPOSED" in MOBIL_parameters
            else LANE_CHANGE_MAX_BRAKING_IMPOSED
        )

    def derive_control_command_from_observation(self, obs_dict):
        """Derive control command from observation.

        Args:
            obs_dict (dict): The observation dictionary.

        Returns:
            dict: The control command.
        """
        if "local" not in obs_dict:
            raise ValueError("No local observation")
        control_command, mode = self.decision(obs_dict["local"])
        return control_command, None

    def decision(self, observation):
        """Vehicle decides next action based on IDM model.

        Args:
            observation (dict): Observation of the vehicle.

        Returns:
            dict, str: The control command. The mode of the decision.
        """
        action = None
        mode = None
        ego_vehicle = observation["Ego"]
        front_vehicle = observation["Lead"]
        # Lateral: MOBIL
        mode = "MOBIL"
        left_gain, right_gain = 0, 0
        left_LC_flag, right_LC_flag = False, False

        if self.MOBIL_lc_flag:
            # see if the vehicle can change lane
            possible_lane_change = []
            (
                possible_lane_change.append(-1)
                if observation["Ego"]["could_drive_adjacent_lane_right"]
                else None
            )
            (
                possible_lane_change.append(1)
                if observation["Ego"]["could_drive_adjacent_lane_left"]
                else None
            )
            # calculate the gain of changing lane and the lane change flag
            for lane_index in possible_lane_change:
                LC_flag, gain = self.mobil_gain(lane_index, observation)
                if LC_flag and gain:
                    if lane_index < 0:
                        right_gain, right_LC_flag = np.clip(gain, 0.0, None), LC_flag
                    elif lane_index > 0:
                        left_gain, left_LC_flag = np.clip(gain, 0.0, None), LC_flag

        if left_LC_flag or right_LC_flag:
            if right_gain > left_gain:
                action = {"lateral": "right", "longitudinal": 0}
                assert right_gain >= self.LANE_CHANGE_MIN_ACC_GAIN
            else:
                action = {"lateral": "left", "longitudinal": 0}
                assert left_gain >= self.LANE_CHANGE_MIN_ACC_GAIN
        # Longitudinal: IDM
        else:
            mode = "IDM"
            if not self.stochastic_acc_flag:
                tmp_acc = self.IDM_acceleration(
                    ego_vehicle=ego_vehicle, front_vehicle=front_vehicle
                )
            else:
                tmp_acc = self.stochastic_IDM_acceleration(
                    ego_vehicle=ego_vehicle, front_vehicle=front_vehicle
                )
            tmp_acc = np.clip(tmp_acc, self.acc_low, self.acc_high)
            if not self.MOBIL_lc_flag:
                action = {"lateral": "SUMO", "longitudinal": tmp_acc}
            else:
                action = {"lateral": "central", "longitudinal": tmp_acc}
        action["type"] = "lon_lat"
        return action, mode

    def IDM_acceleration(self, ego_vehicle=None, front_vehicle=None, distance=None):
        """Compute an acceleration command with the Intelligent Driver Model. The acceleration is chosen so as to:
            - reach a target velocity;
            - maintain a minimum safety distance (and safety time) w.r.t the front vehicle.

        Args:
            ego_vehicle (dict, optional): Information of the vehicle whose desired acceleration is to be computed. It does not have to be an IDM vehicle, which is why this method is a class method. This allows an IDM vehicle to reason about other vehicles behaviors even though they may not IDMs. Defaults to None.
            front_vehicle (dict, optional): Information of the vehicle preceding the ego-vehicle. Defaults to None.
            mode (str, optional): Difference IDM parameters for BV and CAV. Defaults to None.

        Returns:
            float: Acceleration command for the ego-vehicle in m/s^2.
        """
        if not ego_vehicle:
            return 0

        a0 = self.COMFORT_ACC_MAX
        v0 = self.DESIRED_VELOCITY
        delt = self.DELTA
        acceleration = a0 * (1 - np.power(ego_vehicle["velocity"] / v0, delt))
        if front_vehicle is not None:
            r = front_vehicle["distance"] if distance is None else distance
            d = max(1e-5, r)
            acceleration -= a0 * np.power(self.desired_gap(ego_vehicle, front_vehicle) / d, 2)
        return acceleration

    def stochastic_IDM_acceleration(self, ego_vehicle, front_vehicle):
        """Compute an acceleration command with the Intelligent Driver Model. The acceleration is chosen so as to:
            - reach a target velocity;
            - maintain a minimum safety distance (and safety time) w.r.t the front vehicle.
            The acceleration is stochastic and randomly sampled.

        Args:
            ego_vehicle (dict): Information of the vehicle whose desired acceleration is to be computed.
            front_vehicle (dict): Information of the vehicle preceding the ego-vehicle.

        Returns:
            float: Acceleration command for the ego-vehicle in m/s^2.
        """
        tmp_acc = self.IDM_acceleration(ego_vehicle=ego_vehicle, front_vehicle=front_vehicle)
        tmp_acc = np.clip(tmp_acc, self.acc_low, self.acc_high)
        stochastic_IDM_num = (
            int((self.acc_high - self.acc_low) / self.stochastic_IDM_resolution) + 1
        )
        acc_list = np.linspace(self.acc_low, self.acc_high, stochastic_IDM_num)
        acc_possi_list = stats.norm.pdf(acc_list, tmp_acc, 1)
        # Delete possi if smaller than certain threshold
        acc_possi_list = [
            val if val > self.stochastic_IDM_prob_threshold else 0 for val in acc_possi_list
        ]
        assert sum(acc_possi_list) > 0, "The sum of the probability is 0"
        acc_possi_list = acc_possi_list / (sum(acc_possi_list))
        final_acc_idx = np.random.choice(len(acc_list), 1, replace=False, p=acc_possi_list).item()
        final_acc = acc_list[final_acc_idx]
        return final_acc

    def desired_gap(self, ego_vehicle, front_vehicle=None):
        """Compute the desired distance between a vehicle and its leading vehicle.

        Args:
            ego_vehicle (dict): Information of the controlled vehicle.
            front_vehicle (dict, optional): Information of the leading vehicle. Defaults to None.

        Returns:
            float: Desired distance between the two vehicles in m.
        """
        d0 = self.DISTANCE_WANTED
        tau = self.TIME_WANTED
        ab = -self.COMFORT_ACC_MAX * self.COMFORT_ACC_MIN
        dv = ego_vehicle["velocity"] - front_vehicle["velocity"]
        d_star = d0 + max(
            0,
            ego_vehicle["velocity"] * tau + ego_vehicle["velocity"] * dv / (2 * np.sqrt(ab)),
        )
        return d_star

    def mobil_gain(self, lane_index, observation):
        """MOBIL model for the vehicle.

        Args:
            lane_index (int): Lane change direction, -1 represents right, 1 represents left.
            observation (dict): Observation of the vehicle.

        Returns:
            bool, float: Whether it is feasible to do the lane change maneuver. Lane change gain.
        """
        try:
            mingap = traci.vehicle.getMinGap(observation["Ego"]["veh_id"])
        except:
            mingap = 3.28
        gain = None
        ego = observation["Ego"]

        # Is the maneuver unsafe for the new following vehicle?
        if lane_index == 1:
            new_preceding, new_following = (
                observation["LeftLead"],
                observation["LeftFoll"],
            )
        if lane_index == -1:
            new_preceding, new_following = (
                observation["RightLead"],
                observation["RightFoll"],
            )

        # Check whether will crash immediately
        r_new_preceding, r_new_following = 99999, 99999
        if new_preceding:
            r_new_preceding = new_preceding["distance"]
        if new_following:
            r_new_following = new_following["distance"]
        if r_new_preceding <= 0 or r_new_following <= 0:
            return False, gain

        new_following_distance = (
            new_following["distance"] + new_preceding["distance"] + 2 * mingap + self.LENGTH
            if (new_following and new_preceding)
            else 99999
        )
        new_following_a = self.IDM_acceleration(
            ego_vehicle=new_following,
            front_vehicle=new_preceding,
            distance=new_following_distance,
        )
        new_following_pred_a = self.IDM_acceleration(
            ego_vehicle=new_following,
            front_vehicle=ego,
            distance=new_following["distance"] if new_following else 99999,
        )

        old_preceding, old_following = observation["Lead"], observation["Foll"]
        self_pred_a = self.IDM_acceleration(ego_vehicle=ego, front_vehicle=new_preceding)

        # The deceleration of the new following vehicle after the the LC should not be too big (negative)
        if new_following_pred_a < -self.LANE_CHANGE_MAX_BRAKING_IMPOSED:
            return False, gain

        # Is there an acceleration advantage for me and/or my followers to change lane?
        self_a = self.IDM_acceleration(ego_vehicle=ego, front_vehicle=old_preceding)
        old_following_a = self.IDM_acceleration(
            ego_vehicle=old_following,
            front_vehicle=ego,
            distance=old_following["distance"] if old_following else 99999,
        )
        old_following_pred_a_distance = (
            old_following["distance"] + old_preceding["distance"] + 2 * mingap + self.LENGTH
            if (old_following and old_preceding)
            else 99999
        )
        old_following_pred_a = self.IDM_acceleration(
            ego_vehicle=old_following,
            front_vehicle=old_preceding,
            distance=old_following_pred_a_distance,
        )
        gain = (
            self_pred_a
            - self_a
            + self.POLITENESS
            * (new_following_pred_a - new_following_a + old_following_pred_a - old_following_a)
        )
        if gain <= self.LANE_CHANGE_MIN_ACC_GAIN:
            gain = None
            return False, gain

        # All clear, let's go!
        return True, gain



"""This module defines the interface for agent decision models
"""

from abc import ABC


class AgentDecisionModel(ABC):
    """DecisionModel class deal with the control of the vehicle based on observation"""

    def __init__(self):
        """Initialize the agent decision model.
        """
        self._agent = None  # to be assigned from outside
        self.control_log = {}  # This will have the control log result for each controller

    def _reset(self):
        """Reset the agent decision model.
        """
        pass

    def _install(self, agent):
        """Install the agent decision model.

        Args:
            agent (Agent): The agent object.
        """
        self._agent = agent
        pass

    def derive_control_command_from_observation(self, obs_dict):
        """Derive control command from observation.

        Args:
            obs_dict (dict): The observation dictionary.
        
        Returns:
            dict: The control command.
        """
        raise NotImplementedError("Decision model decision function not implemented!")



from abc import ABC, abstractmethod


class VulnerableRoadUserFactory(ABC):
    """
    Basic VulnerableRoadUserFactory class to help build a vru, each vru will contain three major components: sensors, controllers, and decision models.
    Each user who would like to build a customized vehicle should build a son-class and overwrite the vehicle creation method.
    """

    @abstractmethod
    def create_vulnerable_road_user(self, vru_id, simulator):
        raise NotImplementedError("Create vulnerable road user method not implemented!")



from abc import ABC


class InfoExtractor(ABC):
    def __init__(self, env):
        """Initialize the inforomation extractor.

        Args:
            env (BaseEnv): The environment.
        """
        self.env = env

    def add_initialization_info(self):
        """Add initialization information.
        """
        pass

    def get_snapshot_info(self, control_info):
        """Get the snapshot information.

        Args:
            control_info (dict): The control information.
        """
        pass

    def get_terminate_info(self, stop, reason, additional_info):
        """Get the terminate information.

        Args:
            stop (bool): True if the simulation stops.
            reason (str): The reason of the termination.
            additional_info (dict): The additional information.
        """
        pass



"""This module defines the interface for agent controllers
"""

from abc import ABC
import json
from pydantic import BaseModel
from typing import Any, Dict, Optional, Type


class AgentController(ABC):
    params = {}

    def __init__(
        self,
        simulator: Any,
        control_command_schema: Type[BaseModel],
        params: Optional[Dict[str, Any]] = None,
    ):
        """Initialize the agent controller.

        Args:
            simulator (Simulator): The simulator object.
            control_command_schema (BaseModel): The schema of the control command.
            params (Dict, optional): The parameters of the agent controller.
        """
        self._agent = None  # to be assigned from outside
        self.simulator = simulator
        self.control_command_schema = control_command_schema
        if params:
            self.params.update(params)

    def _install(self, agent):
        """Install the agent controller.

        Args:
            agent (Agent): The agent object.
        """
        self._agent = agent
        pass

    def is_command_legal(self, agent_id, control_command):
        """Check if the control command is legal.

        Args:
            agent_id (str): The ID of the agent.
            control_command (dict): The control command.

        Returns:
            bool: True if the control command is legal.
        """
        return True

    def _is_command_legal(self, agent_id, control_command):
        """Inner function to check if the control command is legal.

        Args:
            agent_id (str): The ID of the agent.
            control_command (dict): The control command.

        Returns:
            bool: True if the control command is legal.
        """
        if not isinstance(control_command, self.control_command_schema) and not (
            isinstance(control_command, dict)
            and self.control_command_schema.model_validate_json(json.dumps(control_command))
        ):
            return False
        return self.is_command_legal(agent_id, control_command)

    def execute_control_command(self, agent_id, control_command, obs_dict):
        """Execute the control command.

        Args:
            agent_id (str): The ID of the agent.
            control_command (dict): The control command.
            obs_dict (dict): The observation dictionary.
        """
        pass



import sumolib


class TrafficNet:
    def __init__(self, sumo_net_file_path):
        """Initialize the SUMO Net object.

        Args:
            sumo_net_file_path (str): The path to the sumo network file.
        """
        self.sumo_net_file_path = sumo_net_file_path
        self.sumo_net = sumolib.net.readNet(self.sumo_net_file_path)

    def get_available_lanes_ids(self):
        """Get the available lanes ids in the sumo network.

        Returns:
            list(str object): Possible lanes to insert vehicles.
        """
        return [lane.getID() for lane in self.get_available_lanes()]

    def get_available_lanes(self):
        """Get the available lanes in the sumo network.

        Returns:
            list(SUMO Lane object): Possible lanes to insert vehicles.
        """
        sumo_edges = self.sumo_net.getEdges()
        available_lanes = []
        for edge in sumo_edges:
            for lane in edge.getLanes():
                available_lanes.append(lane)
        return available_lanes



import math
import uuid

from terasim.overlay import traci


def center_coordinate_to_sumo_coordinate(x, y, heading, length=5):
    """Convert the center coordinate to the SUMO coordinate. the input will be a list of {x, y, heading, length}.
    
    Args:
        x (float): x coordinate.
        y (float): y coordinate.
        heading (float): heading.
        length (int, optional): length. Defaults to 5.

    Returns:
        tuple: x and y coordinates.
    """
    x = x + math.cos(heading) * 0.5 * length
    y = y + math.sin(heading) * 0.5 * length
    return x, y


def sumo_coordinate_to_center_coordinate(x, y, heading, length=5):
    """Convert the SUMO coordinate to the center coordinate. the input will be a list of {x, y, heading, length}.
    
    Args:
        x (float): x coordinate.
        y (float): y coordinate.
        heading (float): heading.
        length (int, optional): length. Defaults to 5.

    Returns:
        tuple: x and y coordinates.
    """
    x = x - math.cos(heading) * 0.5 * length
    y = y - math.sin(heading) * 0.5 * length
    return x, y


def sumo_heading_to_orientation(sumo_heading):
    """Convert the SUMO heading to orientation.

    Args:
        sumo_heading (float): SUMO heading.

    Returns:
        float: Orientation.
    """
    radians = math.radians(90 - sumo_heading)
    return math.atan2(math.sin(radians), math.cos(radians))


def orientation_to_sumo_heading(orientation):
    """Convert the orientation to SUMO heading.

    Args:
        orientation (float): Orientation.

    Returns:
        float: SUMO heading.
    """
    degrees = math.degrees(orientation)
    degrees = (degrees + 360) % 360
    degrees = (90 - degrees) % 360
    return degrees


def getLoadedIDList():
    """Get the list of loaded vehicle IDs.

    Returns:
        list(str): A list of vehicle IDs.
    """
    return traci.simulation.getLoadedIDList()


def get_vehicle_route(vehID):
    """Get route of the vehicle.

    Args:
        vehID (str): Vehicle ID.

    Returns:
        list(str): A list of edge ID.
    """
    return traci.vehicle.getRoute(vehID)


def get_vehicle_length(vehID):
    """Get vehicle length.

    Args:
        vehID (str): Vehicle ID.

    Returns:
        float: Vehicle length in m.
    """
    return traci.vehicle.getLength(vehID)


def get_vehicle_width(vehID):
    """Get vehicle width.

    Args:
        vehID (str): Vehicle ID.

    Returns:
        float: Vehicle width in m.
    """
    return traci.vehicle.getWidth(vehID)


def generate_unique_bv_id():
    """Randomly generate an ID of the background vehicle.

    Returns:
        str: ID of the background vehicle.
    """
    return "BV_" + str(uuid.uuid4())


def remap(v, x, y):
    """Remap the value v from the range of x to the range of y.

    Args:
        v (float): Value to be remapped.
        x (list): Original range.
        y (list): Target range.

    Returns:
        float: Remapped value.
    """
    return y[0] + (v - x[0]) * (y[1] - y[0]) / (x[1] - x[0])


def check_equal(x, y, error):
    """Check if x is approximately equal to y considering the given error.

    Args:
        x (float): Parameter 1.
        y (float): Parameter 2.
        error (float): Specified error.

    Returns:
        bool: True is x and y are close enough. Otherwise, False.
    """
    if abs(x - y) <= error:
        return True
    else:
        return False


def get_step_size():
    """Get the step size of the simulation.

    Returns:
        float: Step size in s.
    """
    return traci.simulation.getDeltaT()


def get_time():
    """Get current simulation time in SUMO.

    Returns:
        float: Simulation time in s.
    """
    return traci.simulation.getTime()


def get_vehicle_lateral_lane_position(vehID):
    """Get the lateral offset of the vehicle.

    Args:
        vehID (str): Vehicle ID.

    Returns:
        float: Lateral offset related to the lane's center.
    """
    return traci.vehicle.getLateralLanePosition(vehID)


def cal_dis_with_start_end_speed(v_start, v_end, acc, time_interval=1.0, v_low=20, v_high=40):
    """Calculate the travel distance with start and end speed and acceleration.

    Args:
        v_start (float): Start speed [m/s].
        v_end (float): End speed [m/s].
        acc (float): Acceleration [m/s^2].
        time_interval (float, optional): Time interval [s]. Defaults to 1.0.
        v_low (int, optional): Low speed [m/s]. Defaults to 20.
        v_high (int, optional): High speed [m/s]. Defaults to 40.

    Returns:
        float: Travel distance in the time interval.
    """
    if v_end == v_low or v_end == v_high:
        t_1 = (v_end - v_start) / acc if acc != 0 else 0
        t_2 = time_interval - t_1
        dis = v_start * t_1 + 0.5 * (acc) * (t_1**2) + v_end * t_2
    else:
        dis = ((v_start + v_end) / 2) * time_interval
    return dis


def cal_euclidean_dist(veh1_position=None, veh2_position=None):
    """Calculate Euclidean distance between two vehicles.

    Args:
        veh1_position (tuple, optional): Position of Vehicle 1 [m]. Defaults to None.
        veh2_position (tuple, optional): Position of Vehicle 2 [m]. Defaults to None.

    Raises:
        ValueError: If the position of fewer than two vehicles are provided, raise error.

    Returns:
        float: Euclidean distance between two vehicles [m].
    """
    if veh1_position is None or veh2_position is None:
        raise ValueError("Fewer than two vehicles are provided!")
    veh1_x, veh1_y = veh1_position[0], veh1_position[1]
    veh2_x, veh2_y = veh2_position[0], veh2_position[1]
    return math.sqrt(pow(veh1_x - veh2_x, 2) + pow(veh1_y - veh2_y, 2))


def get_leading_vehicle(vehID, obs_range):
    """Get the information of the leading vehicle.

    Args:
        vehID (str): ID of the ego vehicle.
        obs_range (float): Observation range.

    Returns:
        dict: necessary information of the leading vehicle, including:
            str: ID of the leading vehicle(accessed by 'veh_id'),
            float: Leading vehicle speed (accessed by 'velocity'),
            tuple(float, float): Leading vehicle position in X and Y (accessed by 'position'),
            int: Leading vehicle lane index (accessed by 'lane_index')
            float: Distance between the ego vehicle and the leading vehicle (accessed by 'distance').
    """
    # get leading vehicle information: a list:
    # first element: leader id
    # second element: distance from leading vehicle to ego vehicle
    # (it does not include the minGap of the ego vehicle)
    leader_info = traci.vehicle.getLeader(vehID, dist=obs_range)  # empty leader: None
    if not leader_info:
        return None
    else:
        r = leader_info[1] + traci.vehicle.getMinGap(vehID)
        return get_ego_vehicle(vehID=leader_info[0], obs_range=obs_range, dist=r)


def get_following_vehicle(vehID, obs_range):
    """Get the information of the following vehicle.

    Args:
        vehID (str): ID of the ego vehicle.
        obs_range (float): Observation range.

    Returns:
        dict: necessary information of the following vehicle, including:
            str: ID of the following vehicle(accessed by 'veh_id'),
            float: Following vehicle speed (accessed by 'velocity'),
            tuple(float, float): Following vehicle position in X and Y (accessed by 'position'),
            int: Following vehicle lane index (accessed by 'lane_index')
            float: Distance between the ego vehicle and the following vehicle (accessed by 'distance').
    """
    # get following vehicle information: a list:
    # first element: follower id
    # second element: distance from ego vehicle to following vehicle
    # (it does not include the minGap of the following vehicle)
    follower_info = traci.vehicle.getFollower(vehID, dist=obs_range)  # empty follower: ('',-1)
    if follower_info[1] == -1:
        return None
    else:
        r = follower_info[1] + traci.vehicle.getMinGap(follower_info[0])
        return get_ego_vehicle(vehID=follower_info[0], obs_range=obs_range, dist=r)


def get_neighboring_leading_vehicle(vehID, obs_range, dir):
    """Get the information of the neighboring leading vehicle.

    Args:
        vehID (str): ID of the ego vehicle.
        obs_range (float): Observation range.
        dir (str): Choose from "left" and "right".

    Returns:
        dict: necessary information of the neighboring leading vehicle, including:
            str: ID of the neighboring leading vehicle(accessed by 'veh_id'),
            float: Neighboring leading vehicle speed (accessed by 'velocity'),
            tuple(float, float): Neighboring leading vehicle position in X and Y (accessed by 'position'),
            int: Neighboring leading vehicle lane index (accessed by 'lane_index')
            float: Distance between the ego vehicle and the neighboring leading vehicle (accessed by 'distance').
    """
    # get neighboring leading vehicle information: a list of tuple:
    # first element: leader id
    # second element: distance from leading vehicle to ego vehicle
    # (it does not include the minGap of the ego vehicle)
    if dir == "left":
        leader_info = traci.vehicle.getNeighbors(vehID, 2)  # empty leftleader: len=0
    elif dir == "right":
        leader_info = traci.vehicle.getNeighbors(vehID, 3)  # empty rightleader: len=0
    else:
        raise ValueError("NotKnownDirection when fetching adjacent vehicle information")
    if len(leader_info) == 0:
        return None
    else:
        leader_info_list = [list(item) for item in leader_info]
        for i in range(len(leader_info)):
            leader_info_list[i][1] += traci.vehicle.getMinGap(vehID)
        sorted_leader = sorted(leader_info_list, key=lambda l: l[1])
        closest_leader = sorted_leader[0]
        return get_ego_vehicle(vehID=closest_leader[0], obs_range=obs_range, dist=closest_leader[1])


def get_neighboring_following_vehicle(vehID, obs_range, dir):
    """Get the information of the neighboring following vehicle.

    Args:
        vehID (str): ID of the ego vehicle.
        obs_range (float): Observation range.
        dir (str): Choose from "left" and "right".

    Returns:
        dict: necessary information of the neighboring following vehicle, including:
            str: ID of the neighboring following vehicle(accessed by 'veh_id'),
            float: Neighboring following vehicle speed (accessed by 'velocity'),
            tuple(float, float): Neighboring following vehicle position in X and Y (accessed by 'position'),
            int: Neighboring following vehicle lane index (accessed by 'lane_index')
            float: Distance between the ego vehicle and the neighboring following vehicle (accessed by 'distance').
    """
    # get neighboring following vehicle information: a list of tuple:
    # first element: follower id
    # second element: distance from ego vehicle to following vehicle
    # (it does not include the minGap of the following vehicle)
    if dir == "left":
        follower_info = traci.vehicle.getNeighbors(vehID, 0)  # empty leftfollower: len=0
    elif dir == "right":
        follower_info = traci.vehicle.getNeighbors(vehID, 1)  # empty rightfollower: len=0
    else:
        raise ValueError("NotKnownDirection when fetching adjacent vehicle information")
    if len(follower_info) == 0:
        return None
    else:
        follower_info_list = [list(item) for item in follower_info]
        for i in range(len(follower_info)):
            follower_info_list[i][1] += traci.vehicle.getMinGap(follower_info_list[i][0])
        sorted_follower = sorted(follower_info_list, key=lambda l: l[1])
        closest_follower = sorted_follower[0]
        return get_ego_vehicle(
            vehID=closest_follower[0], obs_range=obs_range, dist=closest_follower[1]
        )


def get_ego_vehicle(vehID, obs_range, dist=0.0):
    """Get the information of the ego vehicle.

    Args:
        vehID (str): ID of the ego vehicle.
        obs_range (float): Observation range.
        dist (float, optional): Distance between two vehicles. Defaults to 0.0.

    Returns:
        dict: Necessary information of the ego vehicle, including:
            str: Vehicle ID (accessed by 'veh_id'),
            float: Vehicle speed (accessed by 'velocity'),
            tuple(float, float): Vehicle position in X and Y (accessed by 'position'),
            int: Vehicle lane index (accessed by 'lane_index')
            float: Distance between the ego vehicle and another vehicle (accessed by 'distance').
    """
    ego_veh = None
    if dist <= obs_range:
        ego_veh = {"veh_id": vehID}
        ego_veh["distance"] = dist
        ego_veh["velocity"] = traci.vehicle.getSpeed(vehID)
        ego_veh["position"] = traci.vehicle.getPosition(vehID)
        ego_veh["position3d"] = traci.vehicle.getPosition3D(vehID)
        ego_veh["heading"] = traci.vehicle.getAngle(vehID)
        ego_veh["edge_id"] = traci.vehicle.getRoadID(vehID)
        ego_veh["lane_id"] = traci.vehicle.getLaneID(vehID)
        ego_veh["lane_index"] = traci.vehicle.getLaneIndex(vehID)
        ego_veh["acceleration"] = traci.vehicle.getAcceleration(vehID)
    return ego_veh


def set_vehicle_speedmode(vehID, speedmode=31):
    """Set the speed mode of the vehicle. This command controls how speeds set with the command setSpeed and slowDown are used. Per default, the vehicle may only drive slower than the speed that is deemed safe by the car following model and it may not exceed the bounds on acceleration and deceleration. Furthermore, vehicles follow the right-of-way rules when approaching an intersection and if necessary they brake hard to avoid driving across a red light.

    Args:
        vehID (str): Vehicle ID.
        speedmode (int, optional): This integer is a bitset (bit0 is the least significant bit) with the following fields. Defaults to 31.
            bit0: Regard safe speed.
            bit1: Regard maximum acceleration.
            bit2: Regard maximum deceleration.
            bit3: Regard right of way at intersections.
            bit4: Brake hard to avoid passing a red light.
    """
    traci.vehicle.setSpeedMode(vehID, speedmode)


def set_vehicle_lanechangemode(vehID, lanechangemode=1621):
    """Sets how lane changing in general and lane changing requests by TraCI are performed.

    Args:
        vehID (str): Vehicle ID.
        lanechangemode (int, optional): If an external change lane command (0x13) command is in conflict with the internal request this is resolved by the current value of the vehicles lane change mode. The given integer is interpreted as a bitset (bit0 is the least significant bit) with the following fields. Defaults to 1621.
        The default lane change mode is 0b011001010101 = 1621 which means that the laneChangeModel may execute all changes unless in conflict with TraCI. Requests from TraCI are handled urgently (with cooperative speed adaptations by the ego vehicle and surrounding traffic) but with full consideration for safety constraints.
        The lane change mode when controlling BV is 0b010001010101 = 1109 which means that the laneChangeModel may execute all changes unless in conflict with TraCI. Requests from TraCI are handled urgently without consideration for safety constraints.
            - bit1, bit0: 00 = do no strategic changes; 01 = do strategic changes if not in conflict with a TraCI request; 10 = do strategic change even if - overriding TraCI request.
            - bit3, bit2: 00 = do no cooperative changes; 01 = do cooperative changes if not in conflict with a TraCI request; 10 = do cooperative change even if overriding TraCI request.
            - bit5, bit4: 00 = do no speed gain changes; 01 = do speed gain changes if not in conflict with a TraCI request; 10 = do speed gain change even if overriding TraCI request.
            - bit7, bit6: 00 = do no right drive changes; 01 = do right drive changes if not in conflict with a TraCI request; 10 = do right drive change even if overriding TraCI request.
            - bit9, bit8:
                00 = do not respect other drivers when following TraCI requests, adapt speed to fulfill request;
                01 = avoid immediate collisions when following a TraCI request, adapt speed to fulfill request;
                10 = respect the speed / brake gaps of others when changing lanes, adapt speed to fulfill request;
                11 = respect the speed / brake gaps of others when changing lanes, no speed adaption.
            - bit11, bit10: 00 = do no sublane changes; 01 = do sublane changes if not in conflict with a TraCI request; 10 = do sublane change even if overriding TraCI request.
    """
    traci.vehicle.setLaneChangeMode(vehID, lanechangemode)


def get_speed_without_traCI(vehID):
    """Get the speed of the vehicle without using TraCI.

    Args:
        vehID (str): Vehicle ID.

    Returns:
        float: Speed of the vehicle without TraCI.
    """
    return traci.vehicle.getSpeedWithoutTraCI(vehID)


def get_speed(vehID):
    """Get the (longitudinal) speed in m/s of the named vehicle within the last step.

    Args:
        vehID (str): Vehicle ID.

    Returns:
        float: Speed of the vehicle within the last step.
    """
    return traci.vehicle.getSpeed(vehID)


def get_next_traffic_light(vehID):
    """Returns the information of the next traffic light the vehicle will encounter, which
    will be a list of upcoming traffic lights [(tlsID, tlsIndex, distance, state), ...].

    Args:
        vehID (str): Vehicle ID.

    Returns:
        float: The distance to the first upcoming traffic light (tlsID, tlsIndex, distance, state).
    """
    tls = traci.vehicle.getNextTLS(vehID)
    if len(tls) == 0:
        return ()
    else:
        return tls[0]


def get_vehicle_speedmode(vehID):
    """Retrieves how the values set by speed (0x40) and slowdown (0x14) shall be treated.

    Args:
        vehID (str): Vehicle ID.

    Returns:
        int: This integer is a bitset (bit0 is the least significant bit) with the following fields.
    """
    return traci.vehicle.getSpeedMode(vehID)


def highlight_vehicle(vehID, duration=-1, color=(255, 0, 0, 255)):
    """Highlight the vehicle with the given ID in the GUI.

    Args:
        vehID (str): Vehicle ID.
        duration (int, optional): Duration of the highlight. Defaults to -1.
        color (tuple, optional): Color of the highlight. Defaults to (255, 0, 0, 255).
    """
    traci.vehicle.highlight(vehID, duration=duration, color=color, alphaMax=0.1)


def get_waiting_time(vehID):
    """Returns the waiting time of the named vehicle in seconds.

    Args:
        vehID (str): Vehicle ID.

    Returns:
        float: Waiting time of the vehicle in seconds.
    """
    return traci.vehicle.getWaitingTime(vehID)


def get_vehicle_lanechangemode(vehID):
    """Returns the lane change mode of the named vehicle.

    Args:
        vehID (str): Vehicle ID.

    Returns:
        int: This integer is a bitset (bit0 is the least significant bit) with the following fields.
    """
    return traci.vehicle.getLaneChangeMode(vehID)


def get_distance(vehID):
    """Returns the distance (in m) to the vehicle ahead.

    Args:
        vehID (str): Vehicle ID.

    Returns:
        float: Distance to the vehicle ahead.
    """
    return traci.vehicle.getDistance(vehID)


def get_vehicle_angle(vehID):
    """Returns the angle in degrees of the named vehicle within the last step.

    Args:
        vehID (str): Vehicle ID.

    Returns:
        float: Angle of the vehicle within the last step.
    """
    return traci.vehicle.getAngle(vehID)



import traci.constants as tc
from addict import Dict

from terasim import utils
from terasim.agent.agent_sensor import AgentSensor
from terasim.overlay import traci
from terasim.simulator import Simulator


class LocalSensor(AgentSensor):
    """
    LocalSensor is a basic sensor that subscribe to some SUMO variables of a vehicle.

    A LocalSensor will maintain a observation, which is a nested dictionary observation.time_stamp
    observation: a dictionary{
        'Ego': {'veh_id': vehicle ID, 'speed': vehicle velocity [m/s], 'position': tuple of X,Y coordinates [m], 'heading': vehicle angle [degree], 'lane_index': lane index of vehicle, 'distance': 0 [m], 'acceleration': m/s^2},
        'Lead'
        'Foll'
        'LeftLead'
        'RightLead'
        'LeftFoll'
        'RightFoll'
    }
    """

    def __init__(self, name="local", **params):
        """Initialize the local sensor for the vehicle.

        Args:
            name (str, optional): The name of the sensor. Defaults to "local".
            params (dict, optional): The parameters of the sensor.
        """
        super().__init__(name, **params)

    def fetch(self):
        """Fetch the vehicle information.

        Returns:
            dict: The vehicle information.
        """
        common = dict(vehID=self._agent.id, obs_range=self._params.obs_range)

        return Dict(
            Ego=LocalSensor.get_ego_vehicle_info(veh_id=self._agent.id),
            Lead=utils.get_leading_vehicle(**common),
            LeftLead=utils.get_neighboring_leading_vehicle(dir="left", **common),
            RightLead=utils.get_neighboring_leading_vehicle(dir="right", **common),
            Foll=utils.get_following_vehicle(**common),
            LeftFoll=utils.get_neighboring_following_vehicle(dir="left", **common),
            RightFoll=utils.get_neighboring_following_vehicle(dir="right", **common),
        )

    @staticmethod
    def get_ego_vehicle_info(veh_id, distance=0.0):
        """Modify the vehicle information into a standard form.

        Args:
            veh_id (str, optional): Vehicle ID. Defaults to None.
            distance (float, optional): Distance from the ego vehicle [m]. Defaults to 0.0.

        Returns:
            dict: Standard form of vehicle information.
        """
        veh_info = Dict(
            veh_id=veh_id,
            velocity=traci.vehicle.getSpeed(veh_id),
            position=traci.vehicle.getPosition(veh_id),
            position3d=traci.vehicle.getPosition3D(veh_id),
            heading=traci.vehicle.getAngle(veh_id),
            edge_id=traci.vehicle.getRoadID(veh_id),
            lane_id=traci.vehicle.getLaneID(veh_id),
            lane_index=traci.vehicle.getLaneIndex(veh_id),
            acceleration=traci.vehicle.getAcceleration(veh_id),
            could_drive_adjacent_lane_left=Simulator.get_vehicle_lane_adjacent(veh_id, 1),
            could_drive_adjacent_lane_right=Simulator.get_vehicle_lane_adjacent(veh_id, -1),
            distance=distance,
            lateral_speed=traci.vehicle.getLateralSpeed(veh_id),
            lateral_offset=traci.vehicle.getLateralLanePosition(veh_id),
        )
        return veh_info



"""
This module re-export sumo and carla for the simulation platform
"""

import logging
import os


# ===== load sumo API, use libsumo by default =====
has_libsumo = False
if os.environ.get("USE_LIBSUMO", True) != "0":
    try:
        import libsumo as traci

        has_libsumo = True
    except ImportError:
        logging.warn("Failed to find libsumo, try to traci instead.")
if not has_libsumo:
    try:
        import traci
    except ImportError:
        logging.error("Unable to find traci!")
        raise


should_profile = os.getenv("ENABLE_PROFILING", "False").lower() in ("true", "1")
if should_profile:
    from line_profiler import LineProfiler

    def profile(func):
        def wrapper(*args, **kwargs):
            profiler = LineProfiler()
            profiler.add_function(func)
            profiler.enable_by_count()
            result = func(*args, **kwargs)
            profiler.print_stats()
            return result

        return wrapper
    
else:

    def profile(func):
        return func



from copy import copy

from terasim.agent.agent import Agent, AgentList


class TrafficLight(Agent):
    DEFAULT_PARAMS = dict(
        agent_type="TrafficLight",
    )


class TrafficLightList(AgentList):
    def __add__(self, another_tls_list):
        """Add two TrafficLightList objects.

        Args:
            another_tls_list (TrafficLightList): Another TrafficLightList object.

        Returns:
            TrafficLightList: The combined TrafficLightList object.
        """
        if not isinstance(another_tls_list, TrafficLightList):
            raise TypeError("TrafficLightList object can only be added to another TrafficLightList")
        tls_list = copy(self)
        keys = self.keys()
        for tls in another_tls_list:
            if tls.id in keys:
                print(
                    f"WARNING: traffic light with same id {tls.id} is added and overwrote the traffic light list"
                )
            tls_list[tls.id] = tls
        return tls_list

    def add_trafficlight(self, tlslist):
        """Add traffic light to the traffic light list.

        Args:
            tlslist (TrafficLightList): List of TrafficLight object or a single TrafficLight object.
        """
        if not isinstance(tlslist, list):
            tlslist = [tlslist]

        for tls in tlslist:
            if tls.id in self.keys():
                print(
                    f"WARNING: traffic light with same id {tls.id} exists and this traffic light is dumped and not overriding the traffic light with same id in the original list"
                )
                continue
            self[tls.id] = tls



from terasim.agent.agent_controller import AgentController
from terasim.simulator import traci
import terasim.utils as utils


class StateController(AgentController):
    params = {}

    def __init__(self, simulator, params=None):
        """Initialize the traffic light state controller.

        Args:
            simulator (Simulator): The simulator object.
            params (dict, optional): The parameters of the state controller.
        """
        super().__init__(simulator, params)
        self.is_busy = False
        self.controlled_duration = 0
        self.step_size = utils.get_step_size()

    def set_traffic_light(self, tlsID, state):
        """Set the traffic light state.

        Args:
            tlsID (str): The ID of the traffic light.
            state (str): The state of the traffic light.
        """
        traci.trafficlight.setRedYellowGreenState(tlsID, state)

    def execute_control_command(self, tls_id, control_command, obs_dict):
        """Execute the control command.

        Args:
            tls_id (str): The ID of the traffic light.
            control_command (dict): The control command.
            obs_dict (dict): The observation dictionary.
        """
        # signal control
        self.set_traffic_light(tls_id, control_command)



class PipelineElement:
    def __init__(self, name, executable, priority=0):
        """Initialize the pipeline element.

        Args:
            name (str): The name of the pipeline element.
            executable (function): The executable function.
            priority (int, optional): The priority of the pipeline element. Defaults to 0.
        """
        self.name = name
        self.executable = executable
        self.priority = priority


class Pipeline(list):
    def __init__(self, name, elements):
        """Initialize the pipeline.

        Args:
            name (str): The name of the pipeline.
            elements (list): The elements of the pipeline.
        """
        self.name = name
        self.extend(elements)
        self.sort(key=lambda e: e.priority)

    def __call__(self, *args, **kwargs):
        """Execute the pipeline.

        Args:
            *args: The arguments of the pipeline.
            **kwargs: The keyword arguments of the pipeline.

        Returns:
            bool: The success of the pipeline execution.
        """
        success = True
        for element in self:
            output = element.executable(*args, **kwargs)

            # The pipeline will report False if one of the pipeline element returned False during execution.
            if output == False:
                success = False
                return success
        return success

    def hook(self, name, executable, priority=None):
        """Hook a new element into the pipeline.

        Args:
            name (str): The name of the pipeline element.
            executable (function): The executable function.
            priority (int, optional): The priority of the pipeline element. Defaults to None.
        """
        element = PipelineElement(name, executable, priority=priority)
        self.extend([element])
        self.sort(key=lambda e: e.priority)



"""This module defines the interface for agents
"""

import addict
from attrs import define
from typing import Any, Dict, Iterable, NewType

from terasim.agent.agent_decision_model import AgentDecisionModel
from terasim.agent.agent_sensor import AgentSensor

AgentId = NewType("AgentId", str)


class AgentType:
    def __init__(self) -> None:
        pass

    @staticmethod
    def default() -> "AgentType":
        return AgentType()


@define
class AgentDepartureInfo:
    time: float = None  # Time step at which the vehicle should enter the network. Defaults to None.
    lane: str = "first"  # Lane on which the vehicle should be inserted. Defaults to 'first'.
    lane_id: str = None  # specific lane id where vehicle should be inserted
    position: str = (
        "base"  # Position at which the vehicle should enter the net. Defaults to 'base'.
    )
    speed: str = "0"  # Speed with which the vehicle should enter the network. Defaults to '0'.


@define
class AgentArrivalInfo:
    lane: str = (
        "current"  # Lane at which the vehicle should leave the network. Defaults to 'current'.
    )
    position: str = (
        "max"  # Position at which the vehicle should leave the network. Defaults to 'max'.
    )
    speed: str = (
        "current"  # Speed with which the vehicle should leave the network. Defaults to 'current'.
    )


@define
class AgentInitialInfo:
    route: str
    type: AgentType = AgentType.default()
    depart: AgentDepartureInfo = AgentDepartureInfo()
    arrive: AgentArrivalInfo = AgentArrivalInfo()

    from_taz: str = (
        ""  # Traffic assignment zone where the vehicle should enter the network. Defaults to ''.
    )
    to_taz: str = (
        ""  # Traffic assignment zones where the vehicle should leave the network. Defaults to ''.
    )
    line: str = ""  # A string specifying the id of a public transport line which can be used when specifying person rides. Defaults to ''.
    person_capacity: int = 0  # Number of all seats of the added vehicle. Defaults to 0.
    person_number: int = 0  # Number of occupied seats when the vehicle is inserted. Defaults to 0.


class Agent:
    """
    A basic class holds the essential information for agents (vehicles, pedestrians, traffic lights, etc.) in the simulator
    """

    DEFAULT_PARAMS = dict(
        agent_type="DefaultAgent",
    )

    def __init__(
        self,
        id: AgentId,
        simulator: Any,
        sensors: Iterable[AgentSensor] = [],
        decision_model=None,
        controller=None,
        **params,
    ):
        """Initialize the agent.

        Args:
            id (AgentId): The ID of the agent.
            simulator (Any): The simulator object.
            sensors (Iterable[AgentSensor], optional): The sensors of the agent. Defaults to [].
            decision_model (AgentDecisionModel, optional): The decision model of the agent. Defaults to None.
            controller (AgentController, optional): The controller of the agent. Defaults to None.
            params (dict, optional): The parameters of the agent. Defaults to {}.
        """
        self._id = id
        self._simulator = simulator
        self._params = addict.Dict(self.DEFAULT_PARAMS)
        self._params.update(params)

        self.sensors: Dict[str, AgentSensor] = {}
        for s in sensors:
            if s.name in sensors:
                raise ValueError("Multiple sensors with the same name!")
            self.sensors[s.name] = s

        if not isinstance(decision_model, AgentDecisionModel):
            raise ValueError("Installing non-decision_model instance as decision_model!")
        self.decision_model = decision_model

        self.controller = controller

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return f"{self._params.agent_type}(id: {self.id})"

    @property
    def id(self):
        return self._id

    @property
    def observation(self):
        return self._fetch_observation()

    @property
    def params(self) -> addict.Dict:
        return self._params

    @property
    def simulator(self):
        return self._simulator

    def _install(self):
        """
        This method is designed to be called after the vehicle exists in the simulator. It installs
        the attaching objects (including sensors, the decision model and controller).
        """
        # install sensors
        for name, sensor in self.sensors.items():
            sensor._install(self)

        # install decision model
        if not isinstance(self.decision_model, AgentDecisionModel):
            raise ValueError("Installing non-decision_model instance as decision_model!")
        self.decision_model._install(self)

        # install controller
        self.controller._install(self)

        # apply params
        # self.simulator.set_vehicle_color(self.id, self.params.properties.color)

    def _fetch_observation(self):
        """Fetch the observation from all sensors.

        Returns:
            dict: The observation dictionary.
        """
        obs_dict = {name: self.sensors[name].observation for name in self.sensors}
        return obs_dict

    def _uninstall(self):
        """This method is designed to be called before the vehicle is removed from the simulator. It uninstalls the attaching sensors.
        """
        # uninstall sensors
        for name, sensor in self.sensors.items():
            sensor._uninstall()

    def apply_control(self, control_command):
        """Apply the control command of given agent.

        Args:
            control_command (dict): The given control command of a specific decision maker.
        """
        obs_dict = self._fetch_observation()
        if type(control_command) == list:
            for c in control_command:
                self.apply_control(c)
        else:
            if self.controller._is_command_legal(self.id, control_command):
                self.controller.execute_control_command(self.id, control_command, obs_dict)
            else:
                # logging.warning(f"Control command {control_command} is not legal for Vehicle {id}")
                pass

    def make_decision(self):
        """Make decision of control command for the agent.

        Returns:
            dict: The control command.
            dict: The information of the decision.
        """
        obs_dict = self._fetch_observation()
        control_command, info = self.decision_model.derive_control_command_from_observation(
            obs_dict
        )
        return control_command, info


class AgentList(dict):
    def __init__(self, d):
        """An agent list that store agents. It derives from a dictionary so that one can call a certain vehicle in O(1) time. Rewrote the iter method so it can iterate as a list.
        """
        super().__init__(d)

    def __iter__(self):
        for k in self.keys():
            yield self[k]



from abc import ABC, abstractmethod
from typing import Union

from loguru import logger

from terasim.agent.agent import AgentDepartureInfo, AgentInitialInfo
from terasim.overlay import traci
from terasim.simulator import Simulator
import terasim.utils as utils
from terasim.vehicle.vehicle import VehicleList


class BaseEnv(ABC):
    def __init__(self, vehicle_factory, info_extractor):
        """Initialize the base testing environment.

        Args:
            vehicle_factory (VehicleFactory): The vehicle factory.
            info_extractor (InfoExtractor): The info extractor.
        """
        self.episode_info = {"start_time": None, "end_time": None}
        self.vehicle_list = VehicleList({})
        self.vehicle_factory = vehicle_factory
        self.info_extractor = info_extractor(self)
        self.simulator: Simulator = None  # to be assigned by the simulator

    ########## Abstract methods that must be overwritten by custom envs ##########

    @abstractmethod
    def on_start(self, ctx) -> bool:
        """Functions to be called when the simulation starts.
        """
        pass

    @abstractmethod
    def on_step(self, ctx) -> Union[bool, dict]:
        """Functions to be called in each simulation step.
        If returned value is boolean, then:
            - True means the simulation shall continue
            - False means the simulation is finished normally.
        If returned value is a dictionary, then:
            - The simulation should be stopped and the related information
              is stored in the dictionary
        """
        pass

    @abstractmethod
    def on_stop(self, ctx) -> bool:
        """Functions to be called when the simulation stops.
        """
        pass

    ########## Utility methods that can be called by custom env ##########

    def add_vehicle_type(
        self,
        type_id: str,
        length: float = 5.0,
        width: float = 2.0,
        height: float = 1.5,
        veh_class: str = "passenger",
        **kwargs,
    ):
        """Add a new vehicle type to the simulation.

        Args:
            type_id (str): The ID of the vehicle type.
            length (float, optional): The length of the vehicle. Defaults to 5.0.
            width (float, optional): The width of the vehicle. Defaults to 2.0.
            height (float, optional): The height of the vehicle. Defaults to 1.5.
            veh_class (str, optional): The class of the vehicle. Defaults to "passenger".
        """
        if type_id in traci.vehicletype.getIDList():
            logger.warning(f"Cannot add new vehicle type: Vehicle type {type_id} already exists.")
        else:
            traci.vehicletype.copy("DEFAULT_VEHTYPE", type_id)
            traci.vehicletype.setLength(type_id, length)
            traci.vehicletype.setWidth(type_id, width)
            traci.vehicletype.setHeight(type_id, height)
            traci.vehicletype.setVehicleClass(type_id, veh_class)

    def add_vehicle(
        self,
        veh_id,
        route_id,
        route=None,
        lane=None,
        lane_id=None,
        position=0,
        speed=-1,
        type_id="DEFAULT_VEHTYPE",
        **kwargs,
    ):
        """Add a new vehicle to the simulation.

        Args:
            veh_id (str): The ID of the vehicle.
            route_id (str): The ID of the route.
            route (list, optional): The route of the vehicle. Defaults to None.
            lane (int, optional): The lane of the vehicle. Defaults to None.
            lane_id (str, optional): The ID of the lane. Defaults to None.
            position (int, optional): The position of the vehicle. Defaults to 0.
            speed (int, optional): The speed of the vehicle. Defaults to -1.
            type_id (str, optional): The ID of the vehicle type. Defaults to "DEFAULT_VEHTYPE".
        """
        assert (
            type(route) == list or route is None
        ), "route should be a list or None, route_id is now used for route id assignment"
        if type_id not in traci.vehicletype.getIDList():
            self.add_vehicle_type(type_id, **kwargs)
        # create the vehicle object from the vehicle factory and add the vehicle to the vehicle list
        vehicle = self._add_vehicle_to_env(veh_id)

        if route_id not in traci.route.getIDList() and route is not None:
            traci.route.add(route_id, route)
        # add the vehicle to the simulators
        self.simulator._add_vehicle_to_sim(
            vehicle,
            AgentInitialInfo(
                route=route_id,
                type=type_id,
                depart=AgentDepartureInfo(
                    position=position, speed=speed, lane=lane, lane_id=lane_id
                ),
            ),
        )

        # setup the attached sensors, controllers, etc
        vehicle._install()

    @property
    def sumo_net(self):
        """Return the sumo network object.

        Raises:
            Exception: If the simulator is not connected yet.
            Exception: If the sumo_net is not initialized yet.

        Returns:
            Net: The sumo network object.
        """
        if self.simulator is None:
            raise Exception("Simulator is not connected yet.")
        if not hasattr(self.simulator, "sumo_net"):
            raise Exception("sumo_net is not initialized yet.")
        return self.simulator.sumo_net

    def remove_vehicle(self, veh_id):
        """Remove a vehicle from the simulation.

        Args:
            veh_id (str): The ID of the vehicle.
        """
        vehicle = self.vehicle_list[veh_id]
        vehicle._uninstall()
        self.simulator._remove_vehicle_from_sim(vehicle)
        self._remove_vehicle_from_env(veh_id)

    ########## These methods are those hooked into the pipelines ##########

    # TODO: remove the simulator arguments in these hooks
    def _start(self, simulator, ctx) -> bool:
        """Start the simulation.

        Args:
            simulator (Simulator): The simulator object.
            ctx (dict): The context.    

        Returns:
            bool: True if the simulation starts successfully.
        """
        # Log down initialization information
        self.episode_info = {"start_time": utils.get_time(), "end_time": None}

        return self.on_start(ctx)

    def _step(self, simulator, ctx) -> bool:
        """Run a simulation step.

        Args:
            simulator (Simulator): The simulator object.
            ctx (dict): The context information.

        Returns:
            bool: True if the simulation should continue.
        """
        # First synchronize the vehicle list
        self._maintain_all_vehicles(ctx)

        # Then call custom env defined step
        step_result = self.on_step(ctx)

        # If custom env requested to stop, log some of the information
        if isinstance(step_result, bool):
            if step_result:
                return True
            else:
                self._request_termination("Simulation ends normally", None)
                return False
        elif isinstance(step_result, dict):
            self._request_termination(step_result["reason"], step_result["info"])
            return False
        else:
            raise TypeError("The output of a step should be a boolean or a dictionary")

    def _stop(self, simulator, ctx):
        """Stop the simulation.

        Args:
            simulator (Simulator): The simulator object.
            ctx (dict): The context information.
        """
        self.on_stop(ctx)

    ########## Other private utility functions that should not be directly called by custom env

    def _maintain_all_vehicles(self, ctx):
        """Maintain the vehicle list based on the departed vehicle list and arrived vehicle list.
        """

        if "terasim_controlled_vehicle_ids" in ctx:
            terasim_controlled_vehicle_ids = (
                ctx["terasim_controlled_vehicle_ids"]
                if isinstance(ctx["terasim_controlled_vehicle_ids"], list)
                else [ctx["terasim_controlled_vehicle_ids"]]
            )
            logger.trace("Using the controlled vehicle list from ctx")
            realtime_vehID_set = set(terasim_controlled_vehicle_ids) & set(
                self.simulator.get_vehID_list()
            )
        else:
            realtime_vehID_set = set(self.simulator.get_vehID_list())

        vehID_set = set(self.vehicle_list.keys())
        # log the difference between the two sets
        logger.trace(f"Realtime vehID set: {realtime_vehID_set}")
        logger.trace(f"Current vehID set: {vehID_set}")
        if vehID_set != realtime_vehID_set:
            for vehID in realtime_vehID_set:
                if vehID not in vehID_set:
                    vehicle = self._add_vehicle_to_env(vehID)
                    vehicle._install()
            for vehID in vehID_set:
                if vehID not in realtime_vehID_set:
                    self._remove_vehicle_from_env(vehID)

    def _add_vehicle_to_env(self, veh_id_list):
        """Add vehicles from veh_id_list.

        Args:
            veh_id_list (list(str)): List of vehicle IDs needed to be inserted.

        Raises:
            ValueError: If one vehicle is neither "BV" nor "AV", it should not enter the network.
        """
        single_input = not isinstance(veh_id_list, list)
        if single_input:
            veh_id_list = [veh_id_list]

        output = []
        for veh_id in veh_id_list:
            veh = self.vehicle_factory.create_vehicle(veh_id, self.simulator)
            self.vehicle_list.add_vehicles(veh)
            output.append(veh)
        return output[0] if single_input else output

    def _remove_vehicle_from_env(self, veh_id_list):
        """Delete vehicles in veh_id_list.

        Args:
            veh_id_list (list(str)): List of vehicle IDs needed to be deleted.

        Raises:
            ValueError: If the vehicle is neither "BV" nor "AV", it shouldn't enter the network.
        """
        if not isinstance(veh_id_list, list):
            veh_id_list = [veh_id_list]
        for veh_id in veh_id_list:
            if veh_id in self.vehicle_list:
                self.vehicle_list.pop(veh_id)._uninstall()

    def _request_termination(self, reason, info):
        """Request the termination of the simulation.

        Args:
            reason (str): The reason of the termination.
            info (dict): The information of the termination.
        """
        self.episode_info["end_time"] = utils.get_time() - utils.get_step_size()
        self.info_extractor.get_terminate_info(True, reason, info)
        self.simulator.running = False



from terasim.agent.agent_decision_model import AgentDecisionModel


class DummyStateDecisionModel(AgentDecisionModel):
    """Dummy decision model:
    This decision model will constantly set the traffic light state as green in all directions
    """

    def derive_control_command_from_observation(self, obs_dict):
        """Derive control command from observation.

        Args:
            obs_dict (dict): The observation dictionary.

        Returns:
            dict: The control command.
        """
        return self.get_decision(), None

    def get_decision(self):
        """Get the decision. Dummy decision is to set the traffic light state as green in all directions.

        Returns:
            str: The decision, which contains the states of all the traffic lights.
        """
        return "ggggggggg"



import addict

from terasim.envs.base_complete import BaseEnvComplete
from terasim.params import AgentType


class EnvTemplateComplete(BaseEnvComplete):
    """This Env provides a basic Env implementation.

    Env developers can derived from this class or build their own implementations directly on BaseEnv
    """

    def on_step(self, ctx):
        """Functions to be called at each simulation step.

        Args:
            ctx (dict): The context information.
        """
        # Make decisions and execute commands
        control_cmds, infos = self.make_decisions(ctx)
        self.execute_control_commands(control_cmds)

        # Simulation stop check
        return self.should_continue_simulation()

    def make_decisions(self, ctx):
        """Make decisions for all vehicles.

        Args:
            ctx (dict): The context information.
        """
        # You can also make decisions for specific vehicles, e.g., only let vehicles near the AV make decisions
        # Cooperative decision making is also possible, e.g., let the AV and the BV make decisions together

        # by default, all vehicles in the vehicle list will make decisions
        if "terasim_controlled_vehicle_ids" in ctx:
            control_command_and_info_veh = {
                veh.id: veh.make_decision()
                for veh in self.vehicle_list
                if veh.id in ctx["terasim_controlled_vehicle_ids"]
            }
        else:
            control_command_and_info_veh = {
                veh.id: veh.make_decision() for veh in self.vehicle_list
            }
        if "terasim_controlled_vulnerable_road_user_ids" in ctx:
            control_command_and_info_vru = {
                vru.id: vru.make_decision()
                for vru in self.vulnerable_road_user_list
                if vru.id in ctx["terasim_controlled_vulnerable_road_user_ids"]
            }
        else:
            control_command_and_info_vru = {
                vru.id: vru.make_decision() for vru in self.vulnerable_road_user_list
            }
        control_command_dict = addict.Dict()
        control_command_dict[AgentType.VEHICLE] = addict.Dict(
            {
                veh_id: command_and_info[0]
                for veh_id, command_and_info in control_command_and_info_veh.items()
            }
        )
        control_command_dict[AgentType.VULNERABLE_ROAD_USER] = addict.Dict(
            {
                vru_id: command_and_info[0]
                for vru_id, command_and_info in control_command_and_info_vru.items()
            }
        )
        info_dict = addict.Dict()
        info_dict[AgentType.VEHICLE] = addict.Dict(
            {
                veh_id: command_and_info[1]
                for veh_id, command_and_info in control_command_and_info_veh.items()
            }
        )
        info_dict[AgentType.VULNERABLE_ROAD_USER] = addict.Dict(
            {
                vru_id: command_and_info[1]
                for vru_id, command_and_info in control_command_and_info_vru.items()
            }
        )
        return control_command_dict, info_dict

    def execute_control_commands(self, control_commands: dict):
        """Execute the control commands of all vehicles and vulnerable road users.
        
        Args:
            control_commands (dict): The control commands for all vehicles and vulnerable road users.
        """
        for veh_id, command in control_commands[AgentType.VEHICLE].items():
            self.vehicle_list[veh_id].apply_control(command)
        for vru_id, command in control_commands[AgentType.VULNERABLE_ROAD_USER].items():
            self.vulnerable_road_user_list[vru_id].apply_control(command)

    def should_continue_simulation(self):
        """Check whether the simulation has ends, return False or Dict (including reason and info) to stop the simulation. Or return True to continue the simulation.

        Returns:
            bool or dict: True if the simulation should continue, False or dict if the simulation should stop
        """
        # By default, the simulation will stop when all vehicles leave the network
        if self.simulator.get_vehicle_min_expected_number() == 0:
            return False

        # You can also define your own termination condition, e.g., when the AV reaches the destination,
        # when collisions between AV and BV happen, etc.
        some_condition = False
        if some_condition:
            return dict(reason="All Vehicles Left", info={})

        # Otherwise return True to continue simulation
        return True



from loguru import logger

from terasim.envs.base import BaseEnv
from terasim.vulnerable_road_user.vulnerable_road_user import VulnerableRoadUserList


class BaseEnvComplete(BaseEnv):
    def __init__(self, vehicle_factory, vulnerable_road_user_factory, info_extractor):
        """Initialize the complete and base testing environment.

        Args:
            vehicle_factory (VehicleFactory): The vehicle factory.
            vulnerable_road_user_factory (VulnerableRoadUserFactory): The vulnerable road user factory.
            info_extractor (InfoExtractor): The info extractor.
        """
        super().__init__(vehicle_factory, info_extractor)
        self.vulnerable_road_user_list = VulnerableRoadUserList({})
        self.vulnerable_road_user_factory = vulnerable_road_user_factory

    def _step(self, simulator, ctx) -> bool:
        """The main step function of the environment.

        Args:
            simulator (Simulator): The simulator object.
            ctx (dict): The context dictionary.

        Returns:
            bool: True if the simulation should continue.
        """
        # First synchronize the vehicle list
        self._maintain_all_vehicles(ctx)
        self._maintain_all_vulnerable_road_users(ctx)

        # Then call custom env defined step
        step_result = self.on_step(ctx)

        # If custom env requested to stop, log some of the information
        if isinstance(step_result, bool):
            if step_result:
                return True
            else:
                self._request_termination("Simulation ends normally", None)
                return False
        elif isinstance(step_result, dict):
            self._request_termination(step_result["reason"], step_result["info"])
            return False
        else:
            raise TypeError("The output of a step should be a boolean or a dictionary")

    ########## Other private utility functions that should not be directly called by custom env

    def _maintain_all_vulnerable_road_users(self, ctx):
        """Maintain the vulnerable road user list based on the departed list and arrived list.
        """
        if "terasim_controlled_vulnerable_road_user_ids" in ctx:
            terasim_controlled_vulnerable_road_user_ids = (
                ctx["terasim_controlled_vulnerable_road_user_ids"]
                if isinstance(ctx["terasim_controlled_vulnerable_road_user_ids"], list)
                else [ctx["terasim_controlled_vulnerable_road_user_ids"]]
            )
            logger.trace("Using the controlled vulnerable road user list from ctx")
            realtime_vruID_set = set(terasim_controlled_vulnerable_road_user_ids) & set(
                self.simulator.get_vruID_list()
            )
        else:
            realtime_vruID_set = set(self.simulator.get_vruID_list())

        vruID_set = set(self.vulnerable_road_user_list.keys())
        # log the difference between the two sets
        logger.trace(f"Realtime vehID set: {realtime_vruID_set}")
        logger.trace(f"Current vehID set: {vruID_set}")
        if vruID_set != realtime_vruID_set:
            for vruID in realtime_vruID_set:
                if vruID not in vruID_set:
                    vru = self._add_vulnerbale_road_user_to_env(vruID)
                    vru._install()
            for vruID in vruID_set:
                if vruID not in realtime_vruID_set:
                    self._remove_vulnerable_road_user_from_env(vruID)

    def _add_vulnerbale_road_user_to_env(self, vru_id_list):
        """Add VRUs from vru_id_list.

        Args:
            vru_id_list (list(str)): List of vru IDs needed to be inserted.

        Raises:
            ValueError: If one vehicle is neither "BV" nor "AV", it should not enter the network.
        """
        single_input = not isinstance(vru_id_list, list)
        if single_input:
            vru_id_list = [vru_id_list]

        output = []
        for vru_id in vru_id_list:
            vru = self.vulnerable_road_user_factory.create_vulnerable_road_user(
                vru_id, self.simulator
            )
            self.vulnerable_road_user_list.add_vulnerable_road_users(vru)
            output.append(vru)
        return output[0] if single_input else output

    def _remove_vulnerable_road_user_from_env(self, vru_id_list):
        """Delete vulnerable road users in vru_id_list.

        Args:
            vru_id_list (list(str)): List of vulnerable road user IDs needed to be deleted.

        Raises:
            ValueError: If the vulnerable road user is neither "BV" nor "AV", it shouldn't enter the network.
        """
        if not isinstance(vru_id_list, list):
            vru_id_list = [vru_id_list]
        for vru_id in vru_id_list:
            if vru_id in self.vulnerable_road_user_list:
                self.vulnerable_road_user_list.pop(vru_id)._uninstall()



import traci.constants as tc

from terasim.agent.agent_sensor import AgentSensor
from terasim.simulator import traci


class EgoStateSensor(AgentSensor):
    """A sensor for reporting basic states (position, speed, heading, etc.)"""

    DEFAULT_PARAMS = dict(
        fields={
            "state": tc.TL_RED_YELLOW_GREEN_STATE,
        }
    )

    def __init__(self, name="ego", **params):
        """Initialize the ego state sensor for the traffic light.

        Args:
            name (str, optional): The name of the sensor. Defaults to "ego".
            params (dict, optional): The parameters of the sensor.
        """
        super().__init__(name, **params)

    def subscribe(self) -> None:
        """Subscribe to the traffic light state.
        """
        tls_id = self._agent.id
        tls_ids = list(self.params.fields.values())
        traci.trafficlight.subscribe(tls_id, varIDs=tls_ids)

    def fetch(self) -> dict:
        """Fetch the traffic light state.

        Returns:
            dict: The traffic light state.
        """
        tls_id = self._agent.id
        sub = traci.trafficlight.getSubscriptionResults(tls_id)
        return {name: sub[id] for name, id in self.params.fields.items()}



"""This module defines the interface for agent sensors
"""

from __future__ import annotations

from abc import ABC, abstractmethod
import addict
from typing import TYPE_CHECKING

from terasim.overlay import traci


if TYPE_CHECKING:
    from terasim.simulator import Simulator


class AgentSensor(ABC):
    DEFAULT_PARAMS = {
        "cache": True,
    }

    def __init__(self, name="base", **params):
        """base sensor initialization

        Args:
            params (dict): all sensor paramters
            name (str, optional): define the name of the sensor. Defaults to "base".
        """
        self._agent = None  # to be assigned from outside
        self._name = name
        self._params = addict.Dict(self.DEFAULT_PARAMS)
        self._params.update(params)
        if self._params.cache:
            self._updated_observation = None
            self._updated_time = None

    def __str__(self) -> str:
        """string method

        Returns:
            str: the name of the sensor
        """
        return self._name

    @property
    def name(self) -> str:
        return self._name

    @property
    def params(self) -> addict.Dict:
        return self._params

    @property
    def is_installed(self) -> bool:
        return self._agent is not None

    @property
    def _simulator(self) -> Simulator:
        assert (
            self.is_installed
        ), "Sensor not installed, you should install sensor before accessing the simulator."
        return self._agent._simulator

    @property
    def observation(self):
        assert (
            self.is_installed
        ), "Sensor not installed, you should install sensor before fetching observations."
        if self._params.cache:
            current_time = traci.simulation.getTime()
            if self._updated_time is None or self._updated_time < current_time:
                self._updated_observation = addict.Dict(self.fetch())
                self._updated_time = current_time
            return self._updated_observation
        else:
            return addict.Dict(self.fetch())

    @abstractmethod
    def fetch(self):
        """Fetch the data from the simulator. For callback-based sensors,
        this method should directly return None.

        Note that this method should not be called by the Env users.
        """
        pass

    def _install(self, agent):
        """Install the sensor to the simulator (but not subscribe to the data yet).

        For SUMO sensors, it will be no-op.
        """
        self._agent = agent

    def _uninstall(self):
        """Uninstall the sensor from the simulator.

        For SUMO sensors, it will be no-op.
        """
        pass



from __future__ import annotations

import math
import time
from pathlib import Path
from typing import Optional

import numpy as np
import sumolib
from loguru import logger
from traci import constants as tc

import terasim
import terasim.utils as utils
from terasim.agent.agent import Agent, AgentInitialInfo

from .overlay import has_libsumo, traci
from .pipeline import Pipeline, PipelineElement


class Context:
    def __init__(self) -> None:
        self.plugins = []


class Simulator(object):
    """
    Simulator deals everything about synchronization of states between SUMO and python script
    """

    def __init__(
        self,
        sumo_config_file_path: str or Path,
        sumo_net_file_path: str or Path,
        num_tries: int = 10,
        gui_flag: bool = False,
        output_path: str or Path = None,
        sumo_output_file_types: list = None,
        step_length: float = None,
        realtime_flag: bool = False,
        additional_sumo_args: str or list = None,
    ):
        """Initialize the simulator.
        
        Args:
            sumo_config_file_path (str or Path): SUMO configuration file path.
            sumo_net_file_path (str or Path): SUMO network file path.
            num_tries (int, optional): Number of tries. Defaults to 10.
            gui_flag (bool, optional): GUI flag. Defaults to False.
            output_path (str or Path, optional): Output path. Defaults to None.
            sumo_output_file_types (list, optional): SUMO output file types. Defaults to None.
            step_length (float, optional): Step length. Defaults to None.
            realtime_flag (bool, optional): Realtime flag. Defaults to False.
            additional_sumo_args (str or list, optional): Additional SUMO arguments. Defaults to None.
        """
        self.sumo_net_file_path = Path(sumo_net_file_path)
        self.sumo_config_file_path = Path(sumo_config_file_path)
        assert self.sumo_net_file_path.exists(), "sumo_net_file_path does not exist"
        assert self.sumo_config_file_path.exists(), "sumo_config_file_path does not exist"
        self.sumo_net = sumolib.net.readNet(str(self.sumo_net_file_path), withInternal=True)

        self.gui_flag = gui_flag
        self.sumo_binary = "sumo-gui" if gui_flag else "sumo"
        self.running = False
        self.num_tries = num_tries
        self.sublane_flag = False
        self.step_length = step_length
        self.realtime_flag = realtime_flag
        assert (
            isinstance(additional_sumo_args, list)
            or isinstance(additional_sumo_args, str)
            or additional_sumo_args is None
        ), "additional_sumo_args should be a list of strings or a string"
        self.additional_sumo_args = (
            [additional_sumo_args]
            if isinstance(additional_sumo_args, str)
            else additional_sumo_args
        )
        if output_path is not None:
            self.output_path = Path(output_path)
            self.output_path.mkdir(parents=True, exist_ok=True)
        else:
            self.output_path = None
            print("Warning: output_path is not specified. No output will be generated.")
        self.sumo_output_file_types = (
            sumo_output_file_types if sumo_output_file_types is not None else []
        )
        self._plugin_list = []
        self.ctx = {}  # context for the execution pipeline

        # pipelines
        self.start_pipeline = Pipeline("start_pipeline", [])  # params: simulator, ctx
        self.step_pipeline = Pipeline(
            "step_pipeline",  # params: simulator, ctx
            [
                PipelineElement(
                    "record_step_start_time",
                    self.record_step_start_time,
                    priority=-10000,
                ),
                PipelineElement("sumo_step", self.sumo_step, priority=10),
                PipelineElement(
                    "compensate_step_end_time",
                    self.compensate_step_end_time,
                    priority=10000,
                ),
            ],
        )
        self.stop_pipeline = Pipeline("stop_pipeline", [])  # params: simulator, ctx
        self._add_vehicle_to_sim = Pipeline(
            "add_vehicle_pipeline",  # params: agent, init_info. If init_info is None, then the initial info should be inferred from traci.
            [PipelineElement("sumo_add", self._add_vehicle_to_sumo)],
        )
        self._remove_vehicle_from_sim = Pipeline(
            "remove_vehicle_pipeline",  # params: agent
            [PipelineElement("sumo_remove", self._remove_vehicle_from_sumo)],
        )

    def bind_env(self, env: terasim.envs.base.BaseEnv):
        """Combine the environment with the simulator

        Args:
            env (BaseEnv): Simulation environment
        """
        self.env = env
        self.env.simulator = self
        self.start_pipeline.hook("env_start", env._start, priority=0)
        self.step_pipeline.hook("env_step", env._step, priority=0)
        self.stop_pipeline.hook("env_stop", env._stop, priority=0)

    def start(self):
        """Start SUMO simulation or initialize environment.
        """
        sumo_cmd = [
            self.sumo_binary,
            "-c",
            str(self.sumo_config_file_path),
        ]
        if self.step_length is not None:
            sumo_cmd += ["--step-length", str(self.step_length)]

        if self.sumo_output_file_types is not None:
            if "traj" in self.sumo_output_file_types:
                filename = str(self.output_path / "traj.xml")
                sumo_cmd += ["--amitran-output", filename]
            if "fcd_all" in self.sumo_output_file_types:
                filename = str(self.output_path / "fcd_all.xml")
                sumo_cmd += ["--fcd-output", filename, "--fcd-output.acceleration"]
            if "fcd" in self.sumo_output_file_types:
                filename = str(self.output_path / "fcd.xml")
                sumo_cmd += [
                    "--fcd-output",
                    filename,
                    "--fcd-output.acceleration",
                    "--device.fcd.explicit",
                    "CAV",
                    "--device.fcd.radius",
                    "200",
                ]
            if "lc" in self.sumo_output_file_types:
                filename = str(self.output_path / "lc.xml")
                sumo_cmd += ["--lanechange-output", filename]
            if "collision" in self.sumo_output_file_types:
                filename = str(self.output_path / "collision.xml")
                sumo_cmd += ["--collision-output", filename]
            if "tripinfo" in self.sumo_output_file_types:
                filename = str(self.output_path / "tripinfo.xml")
                sumo_cmd += [
                    "--tripinfo-output",
                    filename,
                    "--tripinfo-output.write-unfinished",
                ]
        # log file
        if self.output_path is not None:
            sumo_cmd += ["-l", str(self.output_path / "run.log")]
        sumo_cmd += self.additional_sumo_args if self.additional_sumo_args is not None else []
        self.sumo_cmd = sumo_cmd
        if has_libsumo:
            traci.start(sumo_cmd)
        else:
            traci.start(sumo_cmd, numRetries=self.num_tries)

        self.start_pipeline(self, self.ctx)
        self.running = True

    def get_vehicle_route_lanes(self, vehID: str):
        """Get list of lanes in the route of the vehicle.

        Args:
            vehID (str): Vehicle ID.

        Returns:
            list(str): A list of list(sumo lane).
        """
        route = traci.vehicle.getRoute(vehID)
        list_lanes = []
        for edgeID in route:
            sumo_edge = self.sumo_net.getEdge(edgeID)
            list_lanes.append(sumo_edge.getLanes())
        return list_lanes

    def add_plugin(self, plugin):
        """Plugin a plugin to the simulator.

        Args:
            plugin (Plugin): Plugin object.
        """
        self._plugin_list.append(plugin)
        plugin.inject(self, self.ctx)

    @property
    def plugins(self):
        """Get the plugin list.

        Returns:
            list: List of plugins.
        """
        return self._plugin_list

    def step(self):
        """Make a simulation step.
        """
        self.step_pipeline(self, self.ctx)

    def get_all_vehicle_information(self, simulator, ctx):
        """Fetch all vehicle information.

        Args:
            simulator (Simulator): Simulator object.
            ctx (dict): Context information.

        Returns:
            dict: Vehicle information dictionary.
        """
        raise NotImplementedError()

    def sumo_step(self, simulator, ctx):
        """Make a simulation step.
        
        Args:
            simulator (Simulator): Simulator object.
            ctx (dict): Context information.
        """
        traci.simulationStep()

    def record_step_start_time(self, simulator, ctx):
        """Record the starting time of the step.
        
        Args:
            simulator (Simulator): Simulator object.
            ctx (dict): Context information.
        """
        self.step_start_time = time.time()

    def compensate_step_end_time(self, simulator, ctx):
        """Compensate for real world time.
        
        Args:
            simulator (Simulator): Simulator object.
            ctx (dict): Context information.
        """
        if self.realtime_flag:
            step_time = time.time() - self.step_start_time
            sim_step_size = utils.get_step_size()
            if step_time < sim_step_size:
                time.sleep(sim_step_size - step_time)
            else:
                logger.critical(step_time)

    def run(self):
        """Run the specific episode.
        """
        self.start()
        while True:
            self.step()
            if self.running is False:
                break
        self.stop()

    def stop(self):
        """Close SUMO simulation.
        """
        self.stop_pipeline(self, self.ctx)
        traci.close()

    def _add_vehicle_to_sumo(self, veh: Agent, init_info: Optional[AgentInitialInfo]):
        """Generate a vehicle in SUMO network.
        
        Args:
            veh (Agent): Vehicle object.
            init_info (AgentInitialInfo, optional): Initial information of the vehicle. Defaults to None.
        """
        if init_info is None:
            return

        if has_libsumo:
            traci.vehicle.add(
                veh.id,
                init_info.route,
                typeID=init_info.type,
                departSpeed=str(init_info.depart.speed),
            )
        else:
            traci.vehicle.add(
                veh.id,
                init_info.route,
                typeID=init_info.type,
                depart=init_info.depart.time,
                departLane=init_info.depart.lane,
                departPos=init_info.depart.position,
                departSpeed=init_info.depart.speed,
                arrivalLane=init_info.arrive.lane,
                arrivalPos=init_info.arrive.position,
                arrivalSpeed=init_info.arrive.speed,
                fromTaz=init_info.from_taz,
                toTaz=init_info.to_taz,
                line=init_info.line,
                personCapacity=init_info.person_capacity,
                personNumber=init_info.person_number,
            )
        if init_info.depart.lane_id is not None:
            traci.vehicle.moveTo(veh.id, init_info.depart.lane_id, init_info.depart.position)

    def _delete_all_vehicles_in_sumo(self):
        """Delete all vehicles in the network.
        """
        for vehID in traci.vehicle.getIDList():
            traci.vehicle.remove(vehID)

    def track_vehicle_gui(self, vehID="CAV"):
        """Track specific vehicle in GUI.

        Args:
            vehID (str, optional): Vehicle ID. Defaults to "CAV".
        """
        traci.gui.trackVehicle(viewID="View #0", vehID=vehID)

    def get_road_ID(self, vehID):
        """Get ID of the road where the vehicle drives on.

        Args:
            vehID (str): Vehicle ID.

        Returns:
            str: Road ID.
        """
        return traci.vehicle.getRoadID(vehID)

    def changeTarget(self, vehID, edgeID):
        """Change the target edge of the vehicle.

        Args:
            vehID (str): Vehicle ID.
            edgeID (str): Edge ID.
        """
        traci.vehicle.changeTarget(vehID, edgeID)

    def detected_crash(self):
        """Detect the crash happened in the last time step.

        Returns:
            bool: True if a collision happenes in the simulation. False if no collision happens.
        """
        colli = traci.simulation.getCollidingVehiclesIDList()
        return colli

    def detect_vehicle_num(self):
        """Determine the vehicle number in the simulation.

        Returns:
            int: Number of vehicles.
        """
        return traci.simulation.getMinExpectedNumber()

    def get_available_lanes_id(self, edge_id, veh_type="passenger"):
        """Get available lanes for the specific edge and vehicle type.

        Args:
            edge_id (str): Edge ID.
            veh_type (str, optional): Vehicle type. Defaults to 'passenger'.

        Returns:
            list(str): List of lane ID.
        """
        lane_num = self.get_edge_lane_number(edge_id)
        lanes_id = []
        for i in range(lane_num):
            lane_id = edge_id + "_" + str(i)
            if veh_type not in self.get_lane_disallowed(lane_id):
                lanes_id.append(lane_id)
        return lanes_id

    def get_available_lanes(self, edge_id=None):
        """Get the available lanes in the sumo network.

        Args:
            edge_id (str, optional): Edge ID. Defaults to None.

        Returns:
            list(sumo lane object): Possible lanes to insert vehicles
        """
        if edge_id == None:
            sumo_edges = self.sumo_net.getEdges()
        else:
            sumo_edges = [self.sumo_net.getEdge(edge_id)]
        available_lanes = []
        for edge in sumo_edges:
            for lane in edge.getLanes():
                available_lanes.append(lane)
        return available_lanes

    def get_cav_travel_distance(self):
        """Get the travel distance of CAV.

        Returns:
            float: Travel distance.
        """
        return traci.vehicle.getDistance("CAV")

    def get_edge_dist(
        self, first_edge_id, first_lane_position, second_edge_id, second_lane_position
    ):
        """Get distance between two edge position.

        Args:
            first_edge_id (str): Edge ID of the first edge.
            first_lane_position (float): Lane position on the first edge.
            second_edge_id (str): Edge ID of the second edge.
            second_lane_position (float): Lane position on the second edge.

        Returns:
            float: Distance between two positions.
        """
        return traci.simulation.getDistanceRoad(
            first_edge_id,
            first_lane_position,
            second_edge_id,
            second_lane_position,
            True,
        )

    def get_edge_length(self, edgeID):
        """Get the length of the edge.

        Args:
            edgeID (str): Edge ID.

        Returns:
            float: Edge length.
        """
        lane_id = edgeID + "_0"
        lane_length = self.get_lane_length(lane_id)
        return lane_length

    def get_lane_length(self, laneID):
        """Get the lane length.

        Args:
            laneID (str): Lane ID.

        Returns:
            float: Lane length.
        """
        return traci.lane.getLength(laneID)

    def get_lane_width(self, laneID):
        """Get lane width.

        Args:
            laneID (str): Lane ID.

        Returns:
            float: Lane width in m.
        """
        return traci.lane.getWidth(laneID)

    def get_lane_links(self, laneID):
        """Get successor lanes together with priority, open and foe for each link.

        Args:
            laneID (str): Lane ID.

        Returns:
            list: Successor lane information.
        """
        return traci.lane.getLinks(laneID)

    @staticmethod
    def get_vehicle_lane_adjacent(vehID, direction):
        """Get whether the vehicle is allowed to drive on the adjacent lane.

        Args:
            vehID (str): Vehicle ID.
            direction (int): 1 represents left, while -1 represents right.

        Returns:
            bool: Whether the vehicle can drive on the specific lane.
        """
        if direction not in [-1, 1]:
            raise ValueError("Unknown direction input:" + str(direction))
        lane_index = Simulator.get_vehicle_lane_index(vehID)
        new_lane_index = lane_index + direction
        edge_id = Simulator.get_vehicle_roadID(vehID)
        lane_num = Simulator.get_edge_lane_number(edge_id)
        if new_lane_index < 0 or new_lane_index >= lane_num:
            # Adjacent lane does not exist.
            return False
        new_lane_id = edge_id + "_" + str(new_lane_index)
        veh_class = traci.vehicle.getVehicleClass(vehID)
        disallowed = Simulator.get_lane_disallowed(new_lane_id)
        return not veh_class in disallowed

    @staticmethod
    def get_vehicle_lane_index(vehID):
        """Get vehicle lane index.

        Args:
            vehID (str): Vehicle ID.

        Returns:
            int: Lane index.
        """
        return traci.vehicle.getLaneIndex(vehID)

    @staticmethod
    def get_lane_disallowed(laneID):
        """Get disallowed vehicle class of the lane.

        Args:
            laneID (str): Lane ID.

        Returns:
            list(str): Disallowed vehicle class, such as "passenger".
        """
        return traci.lane.getDisallowed(laneID)

    def get_route_edges(self, routeID):
        """Get edges of the route.

        Args:
            routeID (str): Route ID.

        Returns:
            list(str): A list of edge ID.
        """
        return traci.route.getEdges(routeID)

    def _remove_vehicle_from_sumo(self, vehicle):
        """Remove the vehicle from the simulation. Unsubscribe all information as well

        Args:
            vehicle (Vehicle): Vehicle object.
        """
        # unsubscribe the vehicle
        vehID = vehicle.id
        traci.vehicle.unsubscribe(vehID)
        # remove the vehicle from the simulation
        traci.vehicle.remove(vehID)

    def set_vehicle_emegency_deceleration(self, vehID, decel):
        """Set the emergency deceleration of the vehicle.

        Args:
            vehID (str): Vehicle ID.
            decel (float): Deceleration value.
        """
        traci.vehicle.setEmergencyDecel(vehID, decel)

    @staticmethod
    def subscribe_vehicle_ego(vehID):
        """Subscribe to store vehicle's ego information.

        Args:
            vehID (str): Vehicle ID.
        """
        traci.vehicle.subscribe(
            vehID,
            [
                tc.VAR_LENGTH,
                tc.VAR_POSITION,
                tc.VAR_SPEED,
                tc.VAR_LANE_INDEX,
                tc.VAR_ANGLE,
                tc.VAR_POSITION3D,
                tc.VAR_EDGES,
                tc.VAR_LANEPOSITION,
                tc.VAR_LANEPOSITION_LAT,
                tc.VAR_SPEED_LAT,
                tc.VAR_ROAD_ID,
                tc.VAR_ACCELERATION,
            ],
        )

    @staticmethod
    def subscribe_vehicle_surrounding(vehID, max_obs_range=120):
        """Subscribe to store vehicle's ego and surrounding information.

        Args:
            vehID (str): Vehicle ID.
        """
        Simulator.subscribe_vehicle_ego(vehID)
        traci.vehicle.subscribeContext(
            vehID,
            tc.CMD_GET_VEHICLE_VARIABLE,
            max_obs_range,
            [
                tc.VAR_LENGTH,
                tc.VAR_POSITION,
                tc.VAR_SPEED,
                tc.VAR_LANE_INDEX,
                tc.VAR_ANGLE,
                tc.VAR_POSITION3D,
                tc.VAR_ROAD_ID,
                tc.VAR_ACCELERATION,
                tc.VAR_LANE_ID,
            ],
        )
        traci.vehicle.addSubscriptionFilterLanes(
            [-2, -1, 0, 1, 2],
            noOpposite=True,
            downstreamDist=max_obs_range,
            upstreamDist=max_obs_range,
        )

    @staticmethod
    def subscribe_vehicle_all_information(vehID, max_obs_range=120):
        """Subscribe to store vehicle's complete information.

        Args:
            vehID (str): Vehicle ID.
        """
        Simulator.subscribe_vehicle_ego(vehID)
        traci.vehicle.subscribeContext(
            vehID,
            tc.CMD_GET_VEHICLE_VARIABLE,
            max_obs_range,
            [
                tc.VAR_LENGTH,
                tc.VAR_POSITION,
                tc.VAR_SPEED,
                tc.VAR_LANE_INDEX,
                tc.VAR_ANGLE,
                tc.VAR_POSITION3D,
                tc.VAR_EDGES,
                tc.VAR_LANEPOSITION,
                tc.VAR_LANEPOSITION_LAT,
                tc.VAR_SPEED_LAT,
                tc.VAR_ROAD_ID,
                tc.VAR_ACCELERATION,
                tc.VAR_LANE_ID,
            ],
        )
        traci.vehicle.addSubscriptionFilterLanes(
            [-2, -1, 0, 1, 2],
            noOpposite=True,
            downstreamDist=max_obs_range,
            upstreamDist=max_obs_range,
        )

    def unsubscribe_vehicle(self, vehID):
        """Unsubscribe the vehicle information.

        Args:
            vehID (str): Vehicle ID.
        """
        traci.vehicle.unsubscribe(vehID)

    def get_vehicle_context_subscription_results(self, vehID):
        """Get subscription results of the context information.

        Args:
            vehID (str): Vehicle ID.

        Returns:
            dict: Context subscription results.
        """
        return traci.vehicle.getContextSubscriptionResults(vehID)

    def get_vehicle_min_expected_number(self):
        """Get the vehicle number in the simulation plus the number of vehicles waiting to start.

        Returns:
            int: Number of vehicles in the simulation and vehicles waiting to start.
        """
        return traci.simulation.getMinExpectedNumber()

    def get_vehicle_speedmode(self, vehID):
        """Get speed mode of the vehicle.

        Args:
            vehID (str): Vehicle ID.

        Returns:
            int: Speed mode.
        """
        return traci.vehicle.getSpeedMode(vehID)

    def get_colliding_vehicle_number(self):
        """Get the number of cars involved in a collision during the simulation.

        Returns:
            int: Return number of vehicles involved in a collision (typically 2 per collision).
        """
        return traci.simulation.getCollidingVehiclesNumber()

    def get_colliding_vehicles(self):
        """Get the IDs of cars involved in a collision during the simulation.

        Returns:
            list(str): Return the first list of vehicle IDs involved in a collision (typically 2 per collision).
        """
        return traci.simulation.getCollidingVehiclesIDList()

    def get_vehicle_length(self, vehID):
        """Get vehicle length.

        Args:
            vehID (str): Vehicle ID.

        Returns:
            float: Vehicle length in m.
        """
        return traci.vehicle.getLength(vehID)

    def get_vehicle_mingap(self, vehID):
        """Get vehicle minimum gap.

        Args:
            vehID (str): Vehicle ID.

        Returns:
            float: Vehicle minimum gap in m.
        """
        return traci.vehicle.getMinGap(vehID)

    def get_vehicle_acc(self, vehID):
        """Get the acceleration in m/s^2 of the named vehicle within the last step.

        Args:
            vehID (str): Vehicle ID.

        Returns:
            float: Vehicle acceleration [m/s^2].
        """
        return traci.vehicle.getAcceleration(vehID)

    def get_vehicle_maxacc(self, vehID):
        """Get vehicle maximum acceleration.

        Args:
            vehID (str): Vehicle ID.

        Returns:
            float: Maximum acceleration in m/s^2.
        """
        return traci.vehicle.getAccel(vehID)

    def get_vehicle_could_change_lane(self, vehID, direction):
        """Check whehther the vehicle could change lane in the specific direction.

        Args:
            vehID (str): Vehicle ID.
            direction (int): 1 represents "left" and -1 represents "right".

        Returns:
            bool: Whehther the vehicle chould change lane.
        """
        return traci.vehicle.couldChangeLane(vehID, direction)

    def get_vehicle_lane_position(self, vehID):
        """Get the lane position of the vehicle.

        Args:
            vehID (str): Vehicle ID.

        Returns:
            float: Lane position in m.
        """
        return traci.vehicle.getLanePosition(vehID)

    @staticmethod
    def get_edge_lane_number(edgeID):
        """Get lane number of the edge.

        Args:
            edgeID (str): Edge ID.

        Returns:
            int: Lane number.
        """
        return traci.edge.getLaneNumber(edgeID)

    def get_vehicle_lane_number(self, vehID):
        """Get lane number of the edge where the vehicle is driving on.

        Args:
            vehID (str): Vehicle ID.

        Returns:
            int: Lane number.
        """
        return self.get_edge_lane_number(self.get_vehicle_roadID(vehID))

    def get_vehicle_maxdecel(self, vehID):
        """Get vehicle maximum deceleration.

        Args:
            vehID (str): Vehicle ID.

        Returns:
            float: Maximum deceleration.
        """
        return traci.vehicle.getDecel(vehID)

    def get_vehicle_maneuver_pdf(self, vehID):
        """Get vehicle maneuver possibility distribution.

        Args:
            vehID (str): Vehicle ID.

        Returns:
            array(float): Possibility distribution of vehicle maneuver.
        """
        pdf_tuple = traci.vehicle.getNDDProb(vehID)  # tuple of strings
        pdf_array = np.zeros(len(pdf_tuple), dtype=float)  # array of floats
        for i in range(len(pdf_tuple)):
            pdf_array[i] = float(pdf_tuple[i])
        return pdf_array

    @staticmethod
    def get_vehicle_roadID(vehID):
        """Get road ID where the vehicle is driving on.

        Args:
            vehID (str): Vehicle ID.

        Returns:
            str: Road ID.
        """
        return traci.vehicle.getRoadID(vehID)

    def get_vehicle_type(self, vehID):
        """Get vehicle type ID.

        Args:
            vehID (str): Vehicle ID.

        Returns:
            str: Type ID.
        """
        return traci.vehicle.getTypeID(vehID)

    def get_vehicle_laneID(self, vehID):
        """Get lane ID where the vehicle is driving on.

        Args:
            vehID (str): Vehicle ID.

        Returns:
            str: Lane ID.
        """
        return traci.vehicle.getLaneID(vehID)

    def get_vehicle_lane_width(self, vehID):
        """Get lane width where the vehicle is driving on.

        Args:
            vehID (str): Vehicle ID.

        Returns:
            float: Lane width in m.
        """
        laneID = self.get_vehicle_laneID(vehID)
        return self.get_lane_width(laneID)

    def get_vehicle_distance_to_edge(self, veh_id, edge_id, edge_position):
        """Get distance from the vehicle to the edge.

        Args:
            veh_id (str): Vehicle ID.
            edge_id (str): Edge ID.
            edge_position (float): Edge position.

        Returns:
            float: Distance between vehicle and edge.
        """
        second_edge_id = self.get_vehicle_roadID(veh_id)
        first_edge_id = edge_id
        second_lane_position = min(
            self.get_vehicle_lane_position(veh_id), self.get_edge_length(second_edge_id)
        )
        first_lane_position = 0
        return traci.simulation.getDistanceRoad(
            first_edge_id,
            first_lane_position,
            second_edge_id,
            second_lane_position,
            True,
        )

    def get_vehicles_dist_road(self, first_veh_id, second_veh_id):
        """Get the distance between two vehicles along the network.

        Args:
            first_veh_id (str): First vehicle ID.
            second_veh_id (str): Second vehicle ID.

        Returns:
            float: Vehicles distance.
        """

        first_edge_id = self.get_vehicle_roadID(first_veh_id)
        second_edge_id = self.get_vehicle_roadID(second_veh_id)
        first_lane_position = min(
            self.get_vehicle_lane_position(first_veh_id),
            self.get_edge_length(first_edge_id),
        )
        second_lane_position = min(
            self.get_vehicle_lane_position(second_veh_id),
            self.get_edge_length(second_edge_id),
        )
        return traci.simulation.getDistanceRoad(
            first_edge_id,
            first_lane_position,
            second_edge_id,
            second_lane_position,
            True,
        )

    def get_vehicles_dist(self, first_veh_pos, second_veh_pos):
        """Get distance between two vehicles.

        Args:
            first_veh_pos (tuple(float,float)): 2D position of the first vehicle.
            second_veh_pos (tuple(float,float)): 2D position of the second vehicle.

        Returns:
            float: Longitudinal distance between two vehicles.
        """
        return traci.simulation.getDistance2D(
            first_veh_pos[0],
            first_veh_pos[1],
            second_veh_pos[0],
            second_veh_pos[1],
            False,
            True,
        )

    def get_vehicles_relative_lane_index(self, ego_vehID, front_vehID):
        """Get relative lane index for two vehicles.

        Args:
            ego_vehID (str): Ego vehicle ID.
            front_vehID (str): Front vehicle ID.

        Returns:
            int: Relative lane index.
        """
        ego_laneID = self.get_vehicle_laneID(ego_vehID)
        ego_roadID = self.get_vehicle_roadID(ego_vehID)
        laneID = ego_laneID
        roadID = ego_roadID
        front_laneID = self.get_vehicle_laneID(front_vehID)
        front_roadID = self.get_vehicle_roadID(front_vehID)
        if front_roadID[0] == ":":
            links = self.get_lane_links(front_laneID)
            if len(links) > 1:
                print(
                    "WARNING: Can't locate vehicles " + str(ego_vehID) + " and " + str(front_vehID)
                )
                return 3
            front_laneID = links[0][0]
            front_roadID = traci.lane.getEdgeID(front_laneID)
        front_lane_index = int(front_laneID.split("_")[-1])
        it = 0
        while it < 10:
            if front_roadID == roadID:
                lane_index = int(laneID.split("_")[-1])
                return front_lane_index - lane_index
            links = self.get_lane_links(laneID)
            # print(links)
            if len(links) > 1:
                print(
                    "WARNING: Can't locate vehicles " + str(ego_vehID) + " and " + str(front_vehID)
                )
                return 3
            else:
                laneID = links[0][0]
                roadID = traci.lane.getEdgeID(laneID)
            it += 1
        print(
            "WARNING: Can't find relative lane index for vehicles "
            + str(ego_vehID)
            + " and "
            + str(front_vehID)
        )
        return 3

    def get_departed_vehID_list(self):
        """Get ID list of vehicles entering the network in the last time step.

        Returns:
            list(str): List of ID of vehicles entering the network in the last time step.
        """
        return traci.simulation.getDepartedIDList()

    def get_vehicle_type_id(self, veh_id):
        """Get the type ID of the vehicle.

        Args:
            veh_id (str): Vehicle ID.

        Returns:
            str: Type ID.
        """
        return traci.vehicle.getTypeID(veh_id)

    def get_arrived_vehID_list(self):
        """Get ID list of vehicles arriving at the final edge in the last time step.

        Returns:
            list(str): List of ID of vehicles leaving the network in the last time step.
        """
        return traci.simulation.getArrivedIDList()

    def get_vehID_list(self):
        """Get ID list of vehicles currently in the network.

        Returns:
            list(str): List of ID of vehicles in the sumo network in the current time step.
        """
        return traci.vehicle.getIDList()

    def get_vehicle_speed(self, vehID):
        """Get the vehicle speed within the last step.

        Args:
            vehID (str): Vehicle ID.

        Returns:
            float: Vehicle speed in m/s.
        """
        return traci.vehicle.getSpeed(vehID)

    def get_vehicle_lateral_speed(self, vehID):
        """Get the lateral speed of the vehicle.

        Args:
            vehID (str): Vehicle ID.

        Returns:
            float: Later speed of the specified vehicle.
        """
        return traci.vehicle.getLateralSpeed(vehID)

    def get_vehicle_position(self, vehID):
        """Get the position of the vehicle.

        Args:
            vehID (str): Vehicle ID.

        Returns:
            tuple(float, float): Position in X,Y coordinates.
        """
        return traci.vehicle.getPosition(vehID)

    def get_vehicle_lane_number(self, vehID):
        """Get the number of lanes in the edge of the vehicle.

        Args:
            vehID (str): Vehicle ID.

        Returns:
            int: Number of lanes.
        """
        return traci.edge.getLaneNumber(traci.vehicle.getRoadID(vehID))

    def set_zoom(self, zoom):
        """Set the current zoom factor for the given view.

        Args:
            zoom (float): Zoom factor.
        """
        traci.gui.setZoom(viewID="View #0", zoom=zoom)

    def set_vehicle_color(self, vehID, rgb):
        """Set the color of a vehicle to separate it from others.

        Args:
            vehID (str): Vehicle ID.
            rgb (tuple(int, int, int, int)): RGB code of the color, i.e. (255,0,0) for the color red. The fourth component (alpha) is optional.
        """
        traci.vehicle.setColor(vehID, rgb)

    def set_vehicle_max_lateralspeed(self, vehID, lat_max_v):
        """Set the maximum lateral speed of vehicle.

        Args:
            vehID (str): Vehicle ID.
            lat_max_v (float): Maximum lateral speed.
        """
        traci.vehicle.setMaxSpeedLat(vehID, lat_max_v)

    def change_vehicle_speed(self, vehID, acceleration, duration=1.0):
        """Fix the acceleration of a vehicle to be a specified value in the specified duration.

        Args:
            vehID (str): Vehicle ID
            acceleration (float): Specified acceleration of vehicle.
            duration (float, optional): Specified time interval to fix the acceleration in s. Defaults to 1.0.
        """
        # traci.vehicle.slowDown take duration + deltaT to reach the desired speed
        init_speed = traci.vehicle.getSpeed(vehID)
        final_speed = init_speed + acceleration * (utils.get_step_size() + duration)
        if final_speed < 0:
            final_speed = 0
        traci.vehicle.slowDown(vehID, final_speed, duration)

    def _cal_lateral_maxSpeed(self, vehID, lane_width, time=1.0):
        """Calculate the maximum lateral speed for lane change maneuver.

        Args:
            vehID (str): Vehicle ID.
            lane_width (float): Width of the lane.
            time (float, optional): Specified time interval to complete the lane change maneuver in s. Defaults to 1.0.

        Raises:
            ValueError: If the maximum lateral acceleration of the vehicle is too small, it is impossible to complete the lane change maneuver in the specified duration.

        Returns:
            float: Maximum lateral speed aiming to complete the lane change behavior in the specified time duration.
        """
        # accelerate laterally to the maximum lateral speed and maintain
        # v^2 - b*v + c = 0
        lat_acc = float(traci.vehicle.getParameter(vehID, "laneChangeModel.lcAccelLat"))
        b, c = lat_acc * (time), lat_acc * lane_width
        delta_power = b**2 - 4 * c
        if delta_power >= 0:
            lat_max_v = (-math.sqrt(delta_power) + b) / 2
        else:
            raise ValueError("The lateral maximum acceleration is too small.")
        return lat_max_v

    def _cal_lateral_distance(self, vehID, direction):
        """Calculate lateral distance to the target lane for a complete lane change maneuver.

        Args:
            vehID (str): Vehicle ID.
            direction (str): Direction, i.e. "left" and "right".

        Raises:
            ValueError: Unknown lane id.
            ValueError: Unknown direction.

        Returns:
            float: Distance in m.
        """
        origin_lane_id = traci.vehicle.getLaneID(vehID)
        edge_id = traci.vehicle.getRoadID(vehID)
        lane_index = int(origin_lane_id.split("_")[-1])
        origin_lane_width = traci.lane.getWidth(origin_lane_id)
        if direction == "left":
            target_lane_id = edge_id + "_" + str(lane_index + 1)
            try:
                target_lane_width = traci.lane.getWidth(target_lane_id)
            except:
                raise ValueError(
                    "Unknown lane id: " + target_lane_id + " in the lane change maneuver."
                )
            latdist = (origin_lane_width + target_lane_width) / 2
        elif direction == "right":
            target_lane_id = edge_id + "_" + str(lane_index - 1)
            try:
                target_lane_width = traci.lane.getWidth(target_lane_id)
            except:
                raise ValueError(
                    "Unknown lane id: " + target_lane_id + " in the lane change maneuver."
                )
            latdist = -(origin_lane_width + target_lane_width) / 2
        else:
            raise ValueError("Unknown direction for lane change command")
        return latdist

    def change_vehicle_sublane_dist(self, vehID, latdist, duration):
        """Change the lateral position of the vehicle.

        Args:
            vehID (str): Vehicle ID.
            latdist (float): Desired lateral distance.
            duration (float): Change duration.
        """
        lat_max_v = self._cal_lateral_maxSpeed(vehID, abs(latdist), duration)
        traci.vehicle.setMaxSpeedLat(vehID, lat_max_v)
        traci.vehicle.changeSublane(vehID, latdist)

    def change_vehicle_lane(self, vehID, direction, duration=1.0):
        """Force a vehicle to complete the lane change maneuver in the duration.

        Args:
            vehID (str): Vehicle ID.
            direction (str): Choose from "LANE_LEFT" and "LANE_RIGHT".
            duration (float, optional): Specified time interval to complete the lane change behavior. Defaults to 1.0.

        Raises:
            ValueError: Direction is neither "LANE_LEFT" nor "LANE_RIGHT".
        """
        if self.sublane_flag:
            latdist = self._cal_lateral_distance(vehID, direction)
            lat_max_v = self._cal_lateral_maxSpeed(vehID, abs(latdist), duration)
            traci.vehicle.setMaxSpeedLat(vehID, lat_max_v)
            traci.vehicle.changeSublane(vehID, latdist)
        else:
            if direction == "left":
                indexOffset = 1
            elif direction == "right":
                indexOffset = -1
            else:
                raise ValueError("Unknown direction for lane change command")
            traci.vehicle.changeLaneRelative(vehID, indexOffset, utils.get_step_size())

    def change_vehicle_position(
        self, vehID, position, edgeID="", lane=-1, angle=-1073741824.0, keepRoute=1
    ):
        """Move the vehicle to the given coordinates and force it's angle to the given value (for drawing).

        Args:
            vehID (str): Vehicle ID.
            position (tuple(float, float)): The specified x,y coordinates.
            edgeID (str, optional): Edge ID. Defaults to "".
            lane (int, optional): Lane index. Defaults to -1.
            angle (float, optional): angle (float, optional): Specified angle of vehicle. If the angle is set to INVALID_DOUBLE_VALUE, the vehicle assumes the natural angle of the edge on which it is driving. Defaults to -1073741824.0.
            keepRoute (int, optional): If keepRoute is set to 1, the closest position within the existing route is taken. If keepRoute is set to 0, the vehicle may move to any edge in the network but it's route then only consists of that edge. If keepRoute is set to 2, the vehicle has all the freedom of keepRoute=0 but in addition to that may even move outside the road network. Defaults to 1.
        """
        x = position[0]
        y = position[1]
        traci.vehicle.moveToXY(vehID, edgeID, lane, x, y, angle, keepRoute)

    def subscribe_signal(self, tlsID):
        """Subscribe to the specified traffic light.
            TL_BLOCKING_VEHICLES = 37
            TL_COMPLETE_DEFINITION_RYG = 43
            TL_COMPLETE_PROGRAM_RYG = 44
            TL_CONTROLLED_JUNCTIONS = 42
            TL_CONTROLLED_LANES = 38
            TL_CONTROLLED_LINKS = 39
            TL_CURRENT_PHASE = 40
            TL_CURRENT_PROGRAM = 41
            TL_EXTERNAL_STATE = 46
            TL_NEXT_SWITCH = 45
            TL_PHASE_DURATION = 36
            TL_PHASE_INDEX = 34
            TL_PRIORITY_VEHICLES = 49
            TL_PROGRAM = 35
            TL_RED_YELLOW_GREEN_STATE = 32
            TL_RIVAL_VEHICLES = 48

        Args:
            tlsID (str): Signal ID.
        """
        traci.trafficlight.subscribe(
            tlsID,
            [tc.TL_CURRENT_PHASE, tc.TL_PHASE_DURATION, tc.TL_RED_YELLOW_GREEN_STATE],
        )

    def get_signal_information(self, tlsID):
        """Get the subscribed information of traffic signal in the last time step.

        Args:
            tlsID (str): Signal ID.

        Returns:
            dict: Subscribed information of the specified traffic light.
        """
        return traci.trafficlight.getSubscriptionResults(tlsID)

    def get_tlsID_list(self):
        """Get a list of traffic light ID.

        Returns:
            list(str): List of all traffic light in the network.
        """
        return traci.trafficlight.getIDList()

    def get_signal_state(self, tlsID):
        """Returns the named tl's state as a tuple of light definitions from rugGyYoO, for red, yed-yellow, green, yellow, off, where lower case letters mean that the stream has to decelerate.

        Args:
            tlsID (str): Signal ID.

        Returns:
            str: Current state of the specified traffic light.
        """
        return traci.trafficlight.getRedYellowGreenState(tlsID)

    def set_signal_logic(self, tlsID, newlogic):
        """Sets a new program for the given tlsID from a Logic object.

        Args:
            tlsID (str): Signal ID.
            newlogic (Logic): New traffic light logic.
        """
        traci.trafficlight.setProgramLogic(tlsID, newlogic)

    def get_vruID_list(self):
        """Get a list of VRU ID.

        Returns:
            list(str): List of all VRU in the network.
        """
        return traci.person.getIDList()



from terasim.agent.agent_decision_model import AgentDecisionModel


class DummySetSUMOTranformDecisionModel(AgentDecisionModel):
    """Dummy SetSUMOTranform decision model:
    This decision model will constantly move the vehicle to the given x, y coordinates
    """

    def derive_control_command_from_observation(self, obs_dict):
        """Derive control command from observation.

        Args:
            obs_dict (dict): The observation dictionary.

        Returns:
            dict: The control command.
        """
        command = {
            "type": "SetSumoTransform",
            "position": (100, 46),  # x, y
            "velocity": None,  # m/s
            "angle": 0.2,  # rad
        }
        return command, None



from .base import BaseEnv
from .base_complete import BaseEnvComplete
from .template import EnvTemplate
from .template_complete import EnvTemplateComplete
from .template_traffic_light import EnvTrafficLightTemplate



from copy import copy

from terasim.agent.agent import Agent, AgentList


class Vehicle(Agent):
    COLOR_RED = (255, 0, 0)
    COLOR_YELLOW = (255, 255, 0)
    COLOR_BLUE = (0, 0, 255)
    COLOR_GREEN = (0, 255, 0)

    DEFAULT_PARAMS = dict(
        agent_type="Vehicle",
        properties={"color": COLOR_YELLOW},
        initial_info={},
        sync_range=120,  # agents within this range of this vehicle will be synchronized
    )


class VehicleList(AgentList):
    def __add__(self, another_vehicle_list):
        """Add two VehicleList objects.

        Args:
            another_vehicle_list (VehicleList): Another VehicleList object.

        Returns:
            VehicleList: The combined VehicleList object.
        """
        if not isinstance(another_vehicle_list, VehicleList):
            raise TypeError("VehicleList object can only be added to another VehicleList")
        vehicle_list = copy(self)
        keys = self.keys()
        for v in another_vehicle_list:
            if v.id in keys:
                print(
                    f"WARNING: vehicle with same id {v.id} is added and overwrote the vehicle list"
                )
            vehicle_list[v.id] = v
        return vehicle_list

    def add_vehicles(self, vlist):
        """Add vehicles to the vehicle list.

        Args:
            vlist (list(Vehicle)): List of Vehicle object or a single Vehicle object.
        """
        if not isinstance(vlist, list):
            vlist = [vlist]

        for v in vlist:
            if v.id in self.keys():
                print(
                    f"WARNING: vehicle with same id {v.id} exists and this vehicle is dumped and not overriding the vehicle with same id in the original list"
                )
                continue
            self[v.id] = v



from terasim.agent.agent_decision_model import AgentDecisionModel


class SUMOModel(AgentDecisionModel):
    def derive_control_command_from_observation(self, obs_dict):
        """Derive control command from observation.

        Args:
            obs_dict (dict): The observation dictionary.
        """            
        return None, None



from typing import Union

from terasim.envs.base import BaseEnv
from terasim.traffic_light.traffic_light import TrafficLightList


class EnvTrafficLightTemplate(BaseEnv):
    def __init__(self, vehicle_factory, tls_factory, info_extractor):
        """Initialize the base testing environment with control of traffic lights.

        Args:
            vehicle_factory (VehicleFactory): The vehicle factory.
            tls_factory (TrafficLightFactory): The traffic light factory.
            info_extractor (InfoExtractor): The info extractor.
        """
        self.tls_list = TrafficLightList({})
        self.tls_factory = tls_factory
        super().__init__(vehicle_factory, info_extractor)

    def on_start(self, ctx) -> bool:
        """Function to be called when the simulation starts.
        """
        pass

    def on_step(self, ctx) -> Union[bool, dict]:
        """Function to be called at each simulation step.

        Args:
            ctx (dict): The context information.
        """
        # Make decisions and execute commands
        control_cmds, infos = self.make_decisions()
        self.execute_control_commands(control_cmds)

        # Simulation stop c`heck
        return self.should_continue_simulation()

    def on_stop(self, ctx) -> bool:
        """Function to be called when the simulation stops.
        """
        pass

    def make_decisions(self):
        """Make decisions for all vehicles and traffic lights.
        """
        # You can also make decisions for specific vehicles, e.g., only let vehicles near the AV make decisions
        # Cooperative decision making is also possible, e.g., let the AV and the BV make decisions together

        # by default, all vehicles in the vehicle list will make decisions
        control_command_and_info = dict()
        control_command_and_info["veh"] = {veh.id: veh.make_decision() for veh in self.vehicle_list}
        control_command_and_info["tls"] = {tls.id: tls.make_decision() for tls in self.tls_list}
        control_command_dict = dict()
        control_command_dict["veh"] = {
            veh_id: command_and_info[0]
            for veh_id, command_and_info in control_command_and_info["veh"].items()
        }
        control_command_dict["tls"] = {
            tls_id: command_and_info[0]
            for tls_id, command_and_info in control_command_and_info["tls"].items()
        }
        info_dict = dict()
        info_dict["veh"] = {
            veh_id: command_and_info[1]
            for veh_id, command_and_info in control_command_and_info["veh"].items()
        }
        info_dict["tls"] = {
            tls_id: command_and_info[1]
            for tls_id, command_and_info in control_command_and_info["tls"].items()
        }
        return control_command_dict, info_dict

    def execute_control_commands(self, control_commands: dict):
        """Execute the control commands of all vehicles and traffic lights.
        """
        for veh_id, command in control_commands["veh"].items():
            self.vehicle_list[veh_id].apply_control(command)
        for tls_id, command in control_commands["tls"].items():
            self.tls_list[tls_id].apply_control(command)

    def should_continue_simulation(self):
        """Check whether the simulation should end, return False or Dict (including reason and info) to stop the simulation. Or return True to continue the simulation.

        Returns:
            bool: True if the simulation should continue.
        """
        # By default, the simulation will stop when all vehicles leave the network
        if self.simulator.get_vehicle_min_expected_number() == 0:
            return False

        # You can also define your own termination condition, e.g., when the AV reaches the destination,
        # when collisions between AV and BV happen, etc.
        some_condition = False
        if some_condition:
            return dict(reason="All Vehicles Left", info={})

        # Otherwise return True to continue simulation
        return True

    ########## These methods are those hooked into the pipelines ##########

    # TODO: remove the simulator arguments in these hooks
    def _step(self, simulator, ctx) -> bool:
        """The step function that will be called by the simulator.

        Args:
            simulator (Simulator): The simulator object.
            ctx (dict): The context information.

        Returns:
            bool: True if the simulation should continue.
        """
        self._maintain_all_vehicles(ctx)
        self._maintain_all_tls(ctx)
        # Then call custom env defined step
        step_result = self.on_step(ctx)

        # If custom env requested to stop, log some of the information
        if isinstance(step_result, bool):
            if step_result:
                return True
            else:
                self._request_termination("Simulation ends normally", None)
                return False
        elif isinstance(step_result, dict):
            self._request_termination(step_result["reason"], step_result["info"])
            return False
        else:
            raise TypeError("The output of a step should be a boolean or a dictionary")

    ########## Other private utility functions that should not be directly called by custom env

    def _maintain_all_tls(self, ctx):
        """Maintain the traffic light list.
        
        Args:
            ctx (dict): The context information.
        """
        if "terasim_controlled_traffic_light_ids" in ctx:
            realtime_tlsID_set = set(ctx["terasim_controlled_traffic_light_ids"])
        else:
            realtime_tlsID_set = set(self.simulator.get_tlsID_list())

        tlsID_set = set(self.tls_list.keys())
        if tlsID_set != realtime_tlsID_set:
            for tlsID in realtime_tlsID_set:
                if tlsID not in tlsID_set:
                    tls = self._add_tls_to_env(tlsID)
                    tls._install()
            for tlsID in tlsID_set:
                if tlsID not in realtime_tlsID_set:
                    self._remove_tls_from_env(tlsID)

    def _add_tls_to_env(self, tls_id_list):
        """Add traffic lights from tls_id_list.

        Args:
            tls_id_list (list(str)): List of traffic lights IDs needed to be inserted.

        Returns:
            TrafficLight: The traffic light object.
        """
        single_input = not isinstance(tls_id_list, list)
        if single_input:
            tls_id_list = [tls_id_list]

        output = []
        for tls_id in tls_id_list:
            tls = self.tls_factory.create_traffic_light(tls_id, self.simulator)
            self.tls_list.add_trafficlight(tls)
            output.append(tls)
        return output[0] if single_input else output

    def _remove_tls_from_env(self, tls_id_list):
        """Delete traffic lights in tls_id_list.

        Args:
            tls_id_list (list(str)): List of traffic light IDs needed to be deleted.
        """
        if not isinstance(tls_id_list, list):
            tls_id_list = [tls_id_list]
        for tls_id in tls_id_list:
            if tls_id in self.tls_list:
                self.tls_list.pop(tls_id)._uninstall()



from bidict import bidict
import numpy as np

from terasim.agent.agent_decision_model import AgentDecisionModel


class HighwayBaseDecisionModel(AgentDecisionModel):
    longi_safety_buffer, lateral_safety_buffer = 2, 2
    v_low, v_high, r_low, r_high, rr_low, rr_high, acc_low, acc_high = (
        20,
        40,
        0,
        115,
        -10,
        8,
        -4,
        2,
    )
    acc_resolution = 0.2
    LENGTH = 5
    ACTION_STEP = 1.0
    num_acc = int(1 + ((acc_high - acc_low) / acc_resolution))
    CAV_acc_low, CAV_acc_high, CAV_acc_step = -4, 2, 0.2
    num_CAV_acc = int((CAV_acc_high - CAV_acc_low) / CAV_acc_step + 1)
    CAV_acc_to_idx_dic = bidict()
    for i in range(num_CAV_acc):
        CAV_acc_to_idx_dic[
            list(np.arange(CAV_acc_low, CAV_acc_high + CAV_acc_step, CAV_acc_step))[i]
        ] = i
    acc_to_idx_dic = bidict()
    for m in range(num_acc):
        acc_to_idx_dic[list(np.linspace(acc_low, acc_high, num=num_acc))[m]] = m

    def step(self):
        """Store ego vehicle information.
        """
        self.ego_info = self.vehicle.observation.information["Ego"]

    @staticmethod
    def _check_longitudinal_safety(obs, pdf_array, lateral_result=None, CAV_flag=False):
        """Check longitudinal safety for vehicle.

        Args:
            obs (dict): Processed observation of the vehicle.
            pdf_array (list(float)): Old possibility distribution of the maneuvers.
            lateral_result (list(float), optional): Possibility distribution of the lateral maneuvers. Defaults to None.
            CAV_flag (bool, optional): Check whether the vehicle is the CAV. Defaults to False.

        Returns:
            list(float): New possibility distribution of the maneuvers after checking the longitudinal direction.
        """
        ego_info = obs["Ego"]
        f_veh_info = obs["Lead"]
        safety_buffer = HighwayBaseDecisionModel.longi_safety_buffer
        for i in range(len(pdf_array) - 1, -1, -1):
            if CAV_flag:
                acc = HighwayBaseDecisionModel.CAV_acc_to_idx_dic.inverse[i]
            else:
                acc = HighwayBaseDecisionModel.acc_to_idx_dic.inverse[i]
            if f_veh_info is not None:
                rr = f_veh_info["velocity"] - ego_info["velocity"]
                r = f_veh_info["distance"]
                criterion_1 = rr + r + 0.5 * (HighwayBaseDecisionModel.acc_low - acc)
                self_v_2, f_v_2 = max(
                    ego_info["velocity"] + acc, HighwayBaseDecisionModel.v_low
                ), max(
                    (f_veh_info["velocity"] + HighwayBaseDecisionModel.acc_low),
                    HighwayBaseDecisionModel.v_low,
                )
                dist_r = (self_v_2**2 - HighwayBaseDecisionModel.v_low**2) / (
                    2 * abs(HighwayBaseDecisionModel.acc_low)
                )
                dist_f = (f_v_2**2 - HighwayBaseDecisionModel.v_low**2) / (
                    2 * abs(HighwayBaseDecisionModel.acc_low)
                ) + HighwayBaseDecisionModel.v_low * (
                    f_v_2 - self_v_2
                ) / HighwayBaseDecisionModel.acc_low
                criterion_2 = criterion_1 - dist_r + dist_f
                if criterion_1 <= safety_buffer or criterion_2 <= safety_buffer:
                    pdf_array[i] = 0
                else:
                    break

        # Only set the decelerate most when none of lateral is OK.
        if lateral_result is not None:
            lateral_feasible = lateral_result[0] or lateral_result[2]
        else:
            lateral_feasible = False
        if np.sum(pdf_array) == 0 and not lateral_feasible:
            pdf_array[0] = 1 if not CAV_flag else np.exp(-2)
            return pdf_array

        if CAV_flag:
            new_pdf_array = pdf_array
        else:
            new_pdf_array = pdf_array / np.sum(pdf_array)
        return new_pdf_array

    @staticmethod
    def _check_lateral_safety(obs, pdf_array, CAV_flag=False):
        """Check the lateral safety of the vehicle.

        Args:
            obs (dict): Processed information of vehicle observation.
            pdf_array (list): Old possibility distribution of the maneuvers.
            CAV_flag (bool, optional): Check whether the vehicle is the CAV. Defaults to False.

        Returns:
            list: New possibility distribution of the maneuvers after checking the lateral direction.
        """
        CAV_observation = obs
        f0, r0 = CAV_observation["LeftLead"], CAV_observation["LeftFoll"]
        f2, r2 = CAV_observation["RightLead"], CAV_observation["RightFoll"]
        CAV_info = CAV_observation["Ego"]
        lane_change_dir = [0, 2]
        nearby_vehs = [[f0, r0], [f2, r2]]
        safety_buffer = HighwayBaseDecisionModel.lateral_safety_buffer
        ### need to change when considering more than 3 lanes
        if not obs["Ego"]["could_drive_adjacent_lane_right"]:
            pdf_array[2] = 0
        elif not obs["Ego"]["could_drive_adjacent_lane_left"]:
            pdf_array[0] = 0
        for lane_index, nearby_veh in zip(lane_change_dir, nearby_vehs):
            if pdf_array[lane_index] != 0:
                f_veh, r_veh = nearby_veh[0], nearby_veh[1]
                if f_veh is not None:
                    rr = f_veh["velocity"] - CAV_info["velocity"]
                    r = f_veh["distance"]
                    dis_change = (
                        rr * HighwayBaseDecisionModel.ACTION_STEP
                        + 0.5
                        * HighwayBaseDecisionModel.acc_low
                        * (HighwayBaseDecisionModel.ACTION_STEP**2)
                    )
                    r_1 = r + dis_change  # 1s
                    rr_1 = (
                        rr + HighwayBaseDecisionModel.acc_low * HighwayBaseDecisionModel.ACTION_STEP
                    )

                    if r_1 <= safety_buffer or r <= safety_buffer:
                        pdf_array[lane_index] = 0
                    elif rr_1 < 0:
                        self_v_2, f_v_2 = max(
                            CAV_info["velocity"], HighwayBaseDecisionModel.v_low
                        ), max(
                            (f_veh["velocity"] + HighwayBaseDecisionModel.acc_low),
                            HighwayBaseDecisionModel.v_low,
                        )
                        dist_r = (self_v_2**2 - HighwayBaseDecisionModel.v_low**2) / (
                            2 * abs(HighwayBaseDecisionModel.acc_low)
                        )
                        dist_f = (f_v_2**2 - HighwayBaseDecisionModel.v_low**2) / (
                            2 * abs(HighwayBaseDecisionModel.acc_low)
                        ) + HighwayBaseDecisionModel.v_low * (
                            f_v_2 - self_v_2
                        ) / HighwayBaseDecisionModel.acc_low
                        r_2 = r_1 - dist_r + dist_f
                        if r_2 <= safety_buffer:
                            pdf_array[lane_index] = 0
                if r_veh is not None:
                    rr = CAV_info["velocity"] - r_veh["velocity"]
                    r = r_veh["distance"]
                    dis_change = (
                        rr * HighwayBaseDecisionModel.ACTION_STEP
                        - 0.5
                        * HighwayBaseDecisionModel.acc_high
                        * (HighwayBaseDecisionModel.ACTION_STEP**2)
                    )
                    r_1 = r + dis_change
                    rr_1 = (
                        rr
                        - HighwayBaseDecisionModel.acc_high * HighwayBaseDecisionModel.ACTION_STEP
                    )
                    if r_1 <= safety_buffer or r <= safety_buffer:
                        pdf_array[lane_index] = 0
                    elif rr_1 < 0:
                        self_v_2, r_v_2 = min(
                            CAV_info["velocity"], HighwayBaseDecisionModel.v_high
                        ), min(
                            (r_veh["velocity"] + HighwayBaseDecisionModel.acc_high),
                            HighwayBaseDecisionModel.v_high,
                        )
                        dist_r = (r_v_2**2 - HighwayBaseDecisionModel.v_low**2) / (
                            2 * abs(HighwayBaseDecisionModel.acc_low)
                        )
                        dist_f = (self_v_2**2 - HighwayBaseDecisionModel.v_low**2) / (
                            2 * abs(HighwayBaseDecisionModel.acc_low)
                        ) + HighwayBaseDecisionModel.v_low * (
                            -r_v_2 + self_v_2
                        ) / HighwayBaseDecisionModel.acc_low
                        r_2 = r_1 - dist_r + dist_f
                        if r_2 <= safety_buffer:
                            pdf_array[lane_index] = 0
        if np.sum(pdf_array) == 0:
            return np.array([0, 1, 0])

        if CAV_flag:
            new_pdf_array = pdf_array
        else:
            new_pdf_array = pdf_array / np.sum(pdf_array)
        return new_pdf_array



import logging
from typing import Union
from pydantic import BaseModel

from terasim.agent.agent_controller import AgentController
import terasim.utils as utils


class HighEfficiencyCommandSchema(BaseModel):
    longitudinal: Union[float, str]
    lateral: str
    type: str = "lon_lat"


class HighEfficiencyController(AgentController):
    params = {
        "v_high": 40,
        "v_low": 20,
        "acc_duration": 0.1,  # the acceleration duration will be 0.1 second
        "lc_duration": 1,  # the lane change duration will be 1 second
    }

    def __init__(self, simulator, params=None):
        """Initialize the high efficiency controller.

        Args:
            simulator (Simulator): The simulator object.
            params (dict, optional): The parameters of the high efficiency controller. Defaults to None.
        """
        super().__init__(simulator, HighEfficiencyCommandSchema, params)
        self.is_busy = False
        self.controlled_duration = 0
        self.step_size = utils.get_step_size()
        self.lc_step_num = int(self.params["lc_duration"] / self.step_size)

    def is_command_legal(self, veh_id, control_command):
        """Check if the control command is legal.
        
        Args:
            veh_id (str): The ID of the vehicle.
            control_command (dict): The control command.
        
        Returns:
            bool: True if the control command is legal.
        """
        # if control command (dict) does not meet the HighEfficiencyCommand pydantic format, raise an error
        if self.is_busy:
            assert self.controlled_duration >= 0
            if self.controlled_duration == 0:
                self.is_busy = False
            else:
                self.controlled_duration -= 1
                self.controlled_duration = max(self.controlled_duration, 0)
            logging.info("Control command assigned while lane change maneuver not finished")
            return False
        else:
            if control_command[
                "lateral"
            ] == "left" and not self.simulator.get_vehicle_lane_adjacent(veh_id, 1):
                logging.info("Left lane change to a invalid lane")
                return False
            if control_command[
                "lateral"
            ] == "right" and not self.simulator.get_vehicle_lane_adjacent(veh_id, -1):
                logging.info("Right lane change to a invalid lane")
                return False
            return True

    def execute_control_command(self, veh_id, control_command, obs_dict):
        """Vehicle acts based on the input action.

        Args:
            veh_id (str): The ID of the vehicle.
            control_command (dict): The control command.
            obs_dict (dict): The observation dictionary.
        """
        # Remove all control limits from SUMO
        utils.set_vehicle_speedmode(veh_id, 0)

        # Longitudinal control
        controlled_acc = control_command["longitudinal"]
        current_velocity = obs_dict["ego"]["velocity"]
        if current_velocity + controlled_acc > self.params["v_high"]:
            controlled_acc = self.params["v_high"] - current_velocity
        elif current_velocity + controlled_acc < self.params["v_low"]:
            controlled_acc = self.params["v_low"] - current_velocity
        # Lateral control
        if control_command["lateral"] == "SUMO":
            utils.set_vehicle_lanechangemode(veh_id)
            self.simulator.change_vehicle_speed(veh_id, controlled_acc, self.params["acc_duration"])
        else:
            utils.set_vehicle_lanechangemode(veh_id, 0)
            if control_command["lateral"] == "central":
                # current_lane_offset = utils.get_vehicle_lateral_lane_position(veh_id)
                # self.simulator.change_vehicle_sublane_dist(veh_id, -current_lane_offset, self.step_size)
                self.simulator.change_vehicle_speed(
                    veh_id, controlled_acc, self.params["acc_duration"]
                )
            else:
                self.simulator.change_vehicle_lane(
                    veh_id, control_command["lateral"], self.params["lc_duration"]
                )
                self.simulator.change_vehicle_speed(
                    veh_id, controlled_acc, self.params["lc_duration"]
                )
                self.controlled_duration = (
                    self.lc_step_num
                )  # begin counting the lane change maneuver timesteps
                self.is_busy = True



from pydantic import BaseModel
from typing import Tuple

from terasim.agent.agent_controller import AgentController
from terasim.overlay import traci


class SumoMoveCommandSchema(BaseModel):
    position: Tuple[float, float]
    velocity: float
    angle: float
    keepRoute: int = 2
    type: str = "SetSumoTransform"


class SUMOMOVEController(AgentController):
    def __init__(self, simulator):
        """Initialize the SUMO move controller for VRU.

        Args:
            simulator (Simulator): The simulator object.
        """
        super().__init__(simulator, control_command_schema=SumoMoveCommandSchema)

    def execute_control_command(self, vru_id, control_command, obs_dict):
        """Execute the control command.

        Args:
            vru_id (str): The ID of the VRU.
            control_command (dict): The control command.
            obs_dict (dict): The observation dictionary.
        """
        self.set_transform_sumo(
            vru_id,
            control_command["position"],
            control_command["velocity"],
            control_command["angle"],
            control_command["keepRoute"],
            control_command["speedmode"],
        )

    def set_transform_sumo(self, vru_id, position, angle, keepRoute, velocity):
        """Apply the SUMO movePosition command to the VRU.

        Args:
            vru_id (str): ID of the person.
            position (tuple): Position of the person.
            angle (float): Angle of the person.
        """
        # Move vehicle to specific position, speed, and angle without specifying the lane and the edge
        traci.person.moveToXY(
            vru_id,
            edgeID="",
            x=position[0],
            y=position[1],
            angle=angle,
            keepRoute=keepRoute,
        )
        if velocity is not None:
            traci.person.setSpeed(vru_id, velocity)



from copy import copy

from terasim.agent.agent import Agent, AgentList


class VulnerableRoadUser(Agent):
    COLOR_RED = (255, 0, 0)
    COLOR_YELLOW = (255, 255, 0)
    COLOR_BLUE = (0, 0, 255)
    COLOR_GREEN = (0, 255, 0)

    DEFAULT_PARAMS = dict(
        agent_type="VulnerableRoadUser",
        properties={"color": COLOR_BLUE},
        initial_info={},
        sync_range=120,  # agents within this range of this vulnerable_road_user will be synchronized
    )


class VulnerableRoadUserList(AgentList):
    def __add__(self, another_vru_list):
        """Add two VulnerableRoadUserList objects.

        Args:
            another_vru_list (VulnerableRoadUserList): Another VulnerableRoadUserList object.
        
        Returns:
            VulnerableRoadUserList: The combined VulnerableRoadUserList object.
        """
        if not isinstance(another_vru_list, VulnerableRoadUserList):
            raise TypeError(
                "VulnerableRoadUserList object can only be added to another VulnerableRoadUserList"
            )
        vru_list = copy(self)
        keys = self.keys()
        for vru in another_vru_list:
            if vru.id in keys:
                print(
                    f"WARNING: vulnerable road user with same id {vru.id} is added and overwrote the vulnerable road user list"
                )
            vru_list[vru.id] = vru
        return vru_list

    def add_vulnerable_road_users(self, vru_list):
        """Add vrus to the vru list.

        Args:
            vru_list (VulnerableRoadUserList): List of vru object or a single vru object.
        """
        if not isinstance(vru_list, list):
            vru_list = [vru_list]

        for vru in vru_list:
            if vru.id in self.keys():
                print(
                    f"WARNING: vulnerable road user with same id {vru.id} exists and this vulnerable road user is dumped and not overriding the vulnerable road user with same id in the original list"
                )
                continue
            self[vru.id] = vru
